/*
 * Created on Mar 25, 2005
 *
 */
package yaes.application.eel6938.classagents.cxbr;

import java.awt.Point;
import yaes.application.eel6938.game.Constants;
import yaes.application.eel6938.game.objects.Agent;
import yaes.application.eel6938.classagents.CxBRAgent;

/**
 * This is an abstract class is used as the foundation of contexts in CxBR, inherit this class to generate a context
 * This class has a general set of functions that are used by each context.
 * @author Viet Trinh
 *
 */
public abstract class CxBR_Context 
{
	protected Agent _agent;
	protected CxBR_Map _map = null;
	protected String _name = "RandomMoveContext";
	private int _avoid_jiggle = 0;
	
	/**
	 * Constructor
	 * @param name Name of the context
	 * @param a The Agent used for the context
	 * @param m The Context Map used
	 */
	public CxBR_Context(String name, Agent a, CxBR_Map m) 
	{
		this._agent = a;
		this._map = m;
		this._name = name;
	}
	
	/**
	 * Override this function with context logic. All of the action rules should be contained within this function.
	 * This function is called when the context is active.
	 *
	 */
	public abstract void execute();

	/**
	 * Override this function with transition sentinel rule logic. All of the sentinel rules should be contained 
	 * within this function. Universal sentinel rules are called in the mission context {@link CxBR_Engine}.
	 * @return The next active context, it could be this context, it could be another context.
	 */
	public abstract String executeSentinelRules();
	
	/**
	 * Use this function to get the name of the context
	 * @return The name of the context
	 */
	public final String getName() { return this._name; }
	
	/**
	 * Override this fucntion with the color of the context.
	 * @return returns the color set by the context
	 */
	public abstract int[] getColor();
	
	//Do not execute this command unless you are certain you want to move, because it updates the map
	protected final boolean move(int direction, int speed)
	{
		if (this._map.canMove(direction, speed))
		{
			this._agent.move(this._map.getDirection(), speed);
			return true;
		}
		return false;
	}

	//Do not execute this command unless you are certain you want to flee, because it updates the map
	protected final boolean flee(int direction)
	{
		if (this._map.canFlee(direction))
		{
			this._agent.flee(direction);
			return true;
		}
		return false;
	}
	
	/**
	 * Use this function to have the agent move towards a point
	 * @param p The point to move towards
	 * @return true if the agent is on the point
	 */
	protected final boolean _move_towards(Point p)
	{
		int direction, speed;
		if (p.x == 0 && p.y == 0) 
		{
			return true;
		}
		//Avoid agent jiggling towards and away from the food
		if (this._avoid_jiggle > 0)
		{
			//System.out.println("Un-Jiggle");
			this._avoid_jiggle--;
			speed = CxBR_Map.MAX_SPEED;
			direction = this._map.getDirection();
			if (this.move(direction, speed) == false)
			{
				direction = this._map.getClearDirection(direction, speed);
				this.move(direction, speed);
			}
			//Pop the last moves out of the trace Stack
			this._map.getLastMove();
			return false;
		}
		if (Math.abs(p.x) > Math.abs(p.y)) 
		{
			if (p.x > 0) 
			{
				direction = Constants.DirectionEast;
			} 
			else 
			{
				direction = Constants.DirectionWest;
			}
			speed = Math.abs((int) p.x);
		} 
		else 
		{
			if (p.y > 0) 
			{
				direction = Constants.DirectionNorth;
			} 
			else 
			{
				direction = Constants.DirectionSouth;
			}
			speed = Math.abs((int) p.y);
		}
		
		//If I cannot move in primary direction, I'll move in the next clear direction
		if (this.move(direction, speed) == false)
		{
			speed = CxBR_Map.MAX_SPEED;
			direction = this._map.getClearDirection(direction, speed);
			this.move(direction, speed);
			this._avoid_jiggle = CxBR_Map.MAX_JIGGLE_CYCLES;
			this._map.getLastMove();
		}
		return false;
	}
	
	/**
	 * Use this function to find the direction to the point if multiplier == 1, or away from the point if multiplier == -1
	 * @param p
	 * @param multiplier
	 * @return the direction to move
	 */
	private int _point_to_direction(Point p, int multiplier)
	{
		int enemyx = p.x * multiplier;
		int enemyy = p.y * multiplier;
		int direction;
		if (Math.abs(enemyx) > Math.abs(enemyy)) 
		{
			if (enemyx > 0) 
				direction = Constants.DirectionEast;
			else 
				direction = Constants.DirectionWest;
		} 
		else 
		{
			if (enemyy > 0) 
				direction = Constants.DirectionNorth;
			else 
				direction = Constants.DirectionSouth;
		}
		return direction;
	}
	
	/**
	 * Use this function to determine if the agent can explore or not.
	 * @return true if the agent can explore more unexplored land
	 */
	public final boolean canExplore()
	{
		if (this._map.getExplorationDirection() == -1)
			return false;
		return true;
	}
	
	/**
	 * Use this function to determine if the agent can eat or not.
	 * @return true if the agent can eat food
	 */
	public final boolean canEat()
	{
		Point foodpoints[] = this._agent.getSensorFoodList();
		return (foodpoints.length != 0);
	}
	
	/**
	 * Use this function to determine if the agent can mate or not.
	 * @return true if the agent can eat mate
	 */
	public final boolean canMate()
	{
		return (this._agent.canMate() && (Math.random() < CxBRAgent.MATE_PROB));
	}
	
	/**
	 * Use this function to determine if the agent can attack or not. The agent can attack if it's energy level is
	 * higher than the opponents energy level. You are not allowed to attack an agent with the same name.
	 * @param restrictions Set this to true if you want restrictions to be applied to canAttack function, for instance your power level vs. opponent power level
	 * If this is set to false, then the restrictions are lifted
	 * @return The point that can be attacked, or null if there is nothing to attack
	 */
	public final Point canAttack(boolean restrictions)
	{
		Point agents[] = this._agent.getSensorAgentList();
		int enemyE, myE = this._agent.getEnergyLevel();
		if (myE < CxBRAgent.ATTACK_MIN_ENERGY)
			return null;
		for(int i=0;i<agents.length;i++)
		{
			//if (agents[i].x == 0 && agents[i].y == 0)
			//	continue;
			if (this._agent.getName(agents[i]).equals(CxBRAgent.REAL_NAME))
				continue;

			enemyE = this._agent.getEnergyLevel(agents[i]);
			//System.out.println("Energy Equations: " + enemyE + " < " + (myE - CxBRAgent.ATTACK_MARGIN) + "(" + myE + ")");
			if (restrictions == false)
				return agents[i];
			if (enemyE < (myE - CxBRAgent.ATTACK_MARGIN))
				return agents[i];
		}
		return null;
	}
	
	public final int canAvoid()
	{
		Point agents[] = this._agent.getSensorAgentList();
		int enemyE, myE = this._agent.getEnergyLevel();
		for(int i=0;i<agents.length;i++)
		{
			if (this._agent.getName(agents[i]).equals(CxBRAgent.REAL_NAME))
				continue;

			enemyE = this._agent.getEnergyLevel(agents[i]);
			//System.out.println("Energy Equations: " + enemyE + " < " + (myE - CxBRAgent.ATTACK_MARGIN) + "(" + myE + ")");
			if (enemyE >= (myE - CxBRAgent.ATTACK_MARGIN))
			{
				return this._point_to_direction(agents[i], -1);
			}
		}
		return -1;
	}
	
	/**
	 * Use this function to determine if there is any enemy to attack, it can attack anyone it sees that is not friendly.
	 * @return Point where the enemy is located
	 */
	public Point canAttackNoLimit()
	{
		Point agents[] = this._agent.getSensorAgentList();
		for(int i=0;i<agents.length;i++)
		{
			if (this._agent.getName(agents[i]).equals(CxBRAgent.REAL_NAME))
				continue;
			return agents[i];
		}
		return null;
	}	
	
	/**
	 * Use this function to determine if the agent can flee or not. The agent can flee if it has been attacked.
	 * It assumes the neutrality of other agents until they turn and attack.
	 * @param active_context The current active context being executed
	 * @return The direction that the agent can flee in. A -1 will be returned if the agent does not need to flee.
	 */
	public final int canFlee(String active_context)
	{
		boolean canFlee = this._agent.canFlee();
		int enemyE, myE = this._agent.getEnergyLevel();
		int direction;
		Point enemy = null;
		
		//Don't flee while attacking
		if (active_context.equalsIgnoreCase("AttackContext"))
		{
			return -1;
		}
		//If no one attacks, we assume friendliness, no fleeing needed
		if (canFlee == false)
			return -1;
		
		int flee_counter = 0;
		Point agents[] = this._agent.getSensorAgentList();
		for(int i=0;i<agents.length;i++)
		{
			if (this._agent.getName(agents[i]).equals(CxBRAgent.REAL_NAME))
				continue;
				
			enemy = agents[i];
			flee_counter++;
			break;
		}
		
		if (flee_counter == 0)
			return -1;
		
		return this._point_to_direction(enemy, -1);
	}
	
	/**
	 * Use this function to determine if the agent can backtrack or not. 
	 * @return true if the agent can backtrack
	 */
	public final boolean canBackTrack()
	{
		return this._map.canBackTrack();
	}
}
