package yaes.application.marketgrid.clients;

import static yaes.framework.agent.ExtendedMessage.Performative.ACCEPT;

import java.util.ArrayList;
import java.util.List;

import yaes.application.marketgrid.Constants;
import yaes.application.marketgrid.subprotocols.SubProtMatchmakerQuery;
import yaes.application.marketgrid.subprotocols.SubProtProviderExecution;
import yaes.application.marketgrid.subprotocols.SubProtProviderPriceQuery;
import yaes.framework.agent.ExtendedMessage;
import yaes.framework.world.World;
import yaes.ui.text.TextUi;

/**
 * @author Lotzi Boloni
 * 
 * This is a client which has a set of tasks, with certain resource
 * requirements. It executes them on providers, by -negotiating the provider
 * -executing -going to the next task
 * 
 */
public class SerialClient extends Client implements Constants {

    enum ClientStates {
        STATE_NO_TASKS, STATE_PRICE_NEGOTIATE, STATE_PROVIDER_FIND_MATCHMAKER, STATE_EXECUTE
    };

    enum MatchMakerConversationStates {
        STATE_NO_ACTION, STATE_SEND_REQUEST, STATE_WAIT_ANSWER, STATE_RETRY
    };

    enum ProviderConversationStates {
        STATE_NOACTION, STATE_SEND_EXECUTION, STATE_WAIT_EXECUTION_CONFIRM, STATE_WAIT_EXECUTION, STATE_EXECUTION_DONE, STATE_RETRY
    }

    private ClientStates state;
    private MatchMakerConversationStates mstate;
    private ProviderConversationStates pstate;
    private String offeredProvider;
    private String conversationId;
    public List<Task> tasks = new ArrayList<Task>();
    Task currentTask = null;
    private SubProtMatchmakerQuery subprotMatchMakerQuery = new SubProtMatchmakerQuery(
            this);
    private SubProtProviderExecution subprotProviderExecution = new SubProtProviderExecution(this);
    private SubProtProviderPriceQuery subprotProviderPriceQuery = new SubProtProviderPriceQuery(this);
    

    public SerialClient(String name, World world, double resourceRequirements,
            int nrOfTasks, int initMoney) {
        super(name, world, initMoney);
        state = ClientStates.STATE_PROVIDER_FIND_MATCHMAKER;
        mstate = MatchMakerConversationStates.STATE_SEND_REQUEST;
        pstate = ProviderConversationStates.STATE_NOACTION;
        // add three task to be executed
        for (int i = 0; i != nrOfTasks; i++) {
            tasks.add(new Task(resourceRequirements, resourceRequirements, 10, 0));
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see yaes.framework.agent.IAgent#action(yaes.framework.world.World)
     */
    public void action() {
        TextUi.println("Action in "+getName() +" state "+state);
        switch (state) {
        case STATE_NO_TASKS:
            break;
        case STATE_PROVIDER_FIND_MATCHMAKER:
            // looking for the next task
            if (currentTask == null) {
                if (!tasks.isEmpty()) {
                    currentTask = tasks.iterator().next();
                    tasks.remove(currentTask);
                } else {
                    state = ClientStates.STATE_NO_TASKS;
                    break;
                }
            }
            actionMatchMaking(world);
            break;
        case STATE_PRICE_NEGOTIATE:
            actionMatchPriceNegotiate(world);
            break;
        case STATE_EXECUTE:
            actionExecution(world);
            break;
        }
    }

    public void actionMatchMaking(World world) {
        switch (mstate) {
        case STATE_SEND_REQUEST:
        case STATE_RETRY: {
            ExtendedMessage message = subprotMatchMakerQuery
                    .createQueryMessage("MatchMaker_0", currentTask
                            .getRequiredResources());
            send(message);
            mstate = MatchMakerConversationStates.STATE_WAIT_ANSWER;
            break;
        }
        case STATE_WAIT_ANSWER: {
            ExtendedMessage message = getExtendedMessage();
            if (message == null) {
                break;
            }
            // @FIXME: add some kind of acceptance criteria
            offeredProvider = (String) message
                    .getValue(SubProtMatchmakerQuery.PARAM_OFFERED_PROVIDER);
            mstate = MatchMakerConversationStates.STATE_NO_ACTION;
            pstate = ProviderConversationStates.STATE_SEND_EXECUTION;
            state = ClientStates.STATE_PRICE_NEGOTIATE;
            break;
        }
        }
    }

    public void actionMatchPriceNegotiate(World world) {

        switch (pstate) {
        case STATE_NOACTION:
            break;
        case STATE_SEND_EXECUTION: {
            ExtendedMessage message = subprotProviderPriceQuery
                    .createQueryMessage(offeredProvider, currentTask.getRequiredResources());
            send(message);
            pstate = ProviderConversationStates.STATE_WAIT_EXECUTION_CONFIRM;
            break;
        }
        case STATE_WAIT_EXECUTION_CONFIRM: {
            ExtendedMessage message = getExtendedMessage();
            if (message == null) {
                break;
            }
            if (message.getPerformative() == ACCEPT) {
                if (message == null) {
                    break;
                }
                double offeredPrice = ((Double) message.getValue(SubProtProviderPriceQuery.PARAM_ASKING_PRICE))
                        .doubleValue();
                if (isAcceptable(offeredPrice)) {
                    pstate = ProviderConversationStates.STATE_SEND_EXECUTION;
                    state = ClientStates.STATE_EXECUTE;
                } else {
                    // the provider can not find. Go back and retry from the
                    // matchmaker
                    state = ClientStates.STATE_PROVIDER_FIND_MATCHMAKER;
                    mstate = MatchMakerConversationStates.STATE_RETRY;
                    pstate = ProviderConversationStates.STATE_NOACTION;
                }
            } else {
                // the provider can not find. Go back and retry from the
                // matchmaker
                state = ClientStates.STATE_PROVIDER_FIND_MATCHMAKER;
                mstate = MatchMakerConversationStates.STATE_RETRY;
                pstate = ProviderConversationStates.STATE_NOACTION;
            }
            break;
        }
        }
    }

    public void actionExecution(World world) {
        switch (pstate) {
        case STATE_NOACTION:
            break;
        case STATE_SEND_EXECUTION: {
            ExtendedMessage message = subprotProviderExecution.createExecutionRequestMessage(offeredProvider, currentTask.getRequiredResources(), currentTask);
            conversationId = message.getConversation();
            send(message);
            pstate = ProviderConversationStates.STATE_WAIT_EXECUTION_CONFIRM;
            break;
        }
        case STATE_WAIT_EXECUTION_CONFIRM: {
            ExtendedMessage message = getExtendedMessageFromConversation(conversationId);
            if (message == null) {
                break;
            }
            if (message.getSubprotocolMessage().equals(SubProtProviderExecution.MSG_EXECUTION_ACCEPT)) {
                pstate = ProviderConversationStates.STATE_WAIT_EXECUTION;
            } else {
                // the provider can not find. Go back and retry from the
                // matchmaker
                state = ClientStates.STATE_PROVIDER_FIND_MATCHMAKER;
                mstate = MatchMakerConversationStates.STATE_RETRY;
                pstate = ProviderConversationStates.STATE_NOACTION;
            }
            break;
        }
        case STATE_WAIT_EXECUTION: {
            ExtendedMessage message = getExtendedMessageFromConversation(conversationId);
            if (message == null) {
                break;
            }
            if (message.getSubprotocolMessage().equals(SubProtProviderExecution.MSG_EXECUTION_FINISHED)) {
                pstate = ProviderConversationStates.STATE_NOACTION;
                mstate = MatchMakerConversationStates.STATE_SEND_REQUEST;
                state = ClientStates.STATE_PROVIDER_FIND_MATCHMAKER;
                currentTask = null;
            }
            break;
        }
        }

    }

    private boolean isAcceptable(double price) {
        // @FIXME: add some kind of acceptance criteria
        return true;
    }

    public String toString() {
        String temp = "SERIALCLIENT\t\t" + getName() + "\t" + state + " "
                + mstate + " " + pstate + "\tMoney:" + account.getValue()
                + "\tTasks:" + tasks.size();
        return temp;
    }

    @Override
    public boolean isFinished() {
        return tasks.isEmpty() && currentTask == null;
    }
}
