package yaes.application.marketgrid.marketmakers;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

import yaes.application.marketgrid.resource.AllocationDescriptor;
import yaes.application.marketgrid.resource.AllocationRequestParameter;
import yaes.application.marketgrid.resource.IPricingRule;


/**
    Describes chunks of resources as seen from MarketMaker's side. Since data
    need consistency, accessors are used instead having them public;
 */

public class MarketMakerAllocation{

  private String provider;
  private AllocationDescriptor allocatedResource;
  private double availablableQuantity = 0.0;
  private double marketPrice = 0.0;
  private Collection<IPricingRule> pricingRules;

  public MarketMakerAllocation(){
  }

  /**
      @param provider provider identification
      @param allocatedResource resource that marketmaker allocated from provider
   */
  public MarketMakerAllocation(String provider, AllocationDescriptor allocatedResource){
    this.provider = provider;
    this.allocatedResource = allocatedResource.getCopy(false);
    availablableQuantity = allocatedResource.quantity;
  }
  /**
      @return true if resource matches request
   */
  public boolean isMatch(Collection<AllocationRequestParameter> resourceParameters){
    return
      AllocationRequestParameter.matches(resourceParameters,
                                         allocatedResource.resourceParameters);
  }
  /**
      @return provider
   */
  public String getProvider(){
    return provider;
  }

  /**
      @return availablable quantity
   */
  public double getAvailableQuantity(){
    return availablableQuantity;
  }

  /**
      Allocates resource
      #param quantity
   */
  public double allocateResource(double quantity){
    if(quantity > availablableQuantity)
      return 0.0;
    availablableQuantity -= quantity;
    return quantity;
  }

  /**
      Releases resource  (if applicable)
      #param quantity
   */
  public void releaseResource(double quantity){
    availablableQuantity += quantity;

  }
  /**
      Merge newly purchased resource. Sets unit price as average price based
      on total values of merging quantities
      @param purchase 
   */
  public void mergeResource(AllocationDescriptor purchase){
    availablableQuantity += purchase.quantity;
    double value = allocatedResource.expectedUnitPrice * allocatedResource.quantity +
            purchase.expectedUnitPrice * purchase.quantity;
    double average_unit_price = value / (purchase.quantity + allocatedResource.quantity);
    allocatedResource.expectedUnitPrice = average_unit_price;
    allocatedResource.quantity += purchase.quantity;
  }
  /**
      Pricing rules as seen from MarketMakers side
      @return pricing Rules
   */
  public Collection<IPricingRule> getPricingRules(){
    return pricingRules;
  }
  /**
      @param pricingRules set of rules. overrides any existing set
   */
  public void setPricingRules(Collection<IPricingRule> pricingRules){
    this.pricingRules = pricingRules;
  }

  /**
      @param pricingRule rule to be added to existing set
   */
  public void addPricingRule(IPricingRule pricingRule){
    if(pricingRules == null)
      pricingRules = new ArrayList<IPricingRule>();
    pricingRules.add(pricingRule);
  }
  /**
      @return purchase cost (price at provider)
   */
  public double getAllocatePrice(){
    return allocatedResource.expectedUnitPrice;
  }

}

