package yaes.application.marketgrid.marketmakers;

import java.util.ArrayList;
import java.util.Collection;

import yaes.application.marketgrid.MarketMaker;
import yaes.application.marketgrid.clients.Task;
import yaes.application.marketgrid.brokers.Offer;
import yaes.application.marketgrid.resource.AllocationDescriptor;
import yaes.application.marketgrid.subprotocols.SubProtMatchmakerQuery;
import yaes.application.marketgrid.subprotocols.SubProtProviderExecution;
import yaes.application.marketgrid.subprotocols.SubProtProviderPriceQuery;
import yaes.application.marketgrid.subprotocols.SubProtProviderWholeSale;
import yaes.application.marketgrid.subprotocols.SubProtProviderCommited;

import yaes.framework.agent.ExtendedMessage;
import yaes.ui.text.TextUi;

public class MarketMakerConversation {

    enum MarketMakerConversationState {
        NO_ACTION,
        // client (broker) requests
        STATE_PRICE_REQUEST,
        STATE_RESOURCE_REQUEST,
        STATE_RESOURCE_EXECUTE_RESPONSE,
        // purchase conversation
        STATE_BUY_RESOURCES, STATE_BUY_MATCH, STATE_BUY_ASK_PRICE, STATE_BUY_ALLOCATE
    }

    private String id = null;
    private MarketMaker marketMaker;
    private String currentClient;
    private MarketMakerConversationState state;
    private AllocationDescriptor requestedResource;
    private SubProtProviderExecution subProtProviderExecution;
    private SubProtProviderPriceQuery subProtProviderPriceQuery;
    private SubProtMatchmakerQuery subProtMatchmakerQuery;
    private SubProtProviderWholeSale subProtProviderWholeSale;
    private double currentInvestment = 0.0;
    private Collection<String> offeredProviders;
    private int answers_count = 0;
    private ArrayList<Offer> providersOffers;

    public MarketMakerConversation(MarketMaker marketMaker,
            ExtendedMessage message) {
        this.marketMaker = marketMaker;
        this.id = message.getConversation();
        subProtProviderPriceQuery = new SubProtProviderPriceQuery(marketMaker);
        subProtMatchmakerQuery = new SubProtMatchmakerQuery(marketMaker);
        subProtProviderWholeSale = new SubProtProviderWholeSale(marketMaker);
        this.currentClient = message.getSender();
        if (message.getValue(SubProtProviderPriceQuery.PARAM_ASKING_PRICE) != null) {
            state = MarketMakerConversationState.STATE_PRICE_REQUEST;
            requestedResource = (AllocationDescriptor) message
                    .getValue(SubProtProviderPriceQuery.PARAM_ASKING_PRICE);
        } else if (message
                .getValue(SubProtProviderExecution.PARAM_REQUESTED_RESOURCES) != null) {
            state = MarketMakerConversationState.STATE_RESOURCE_REQUEST;
            requestedResource = (AllocationDescriptor) message
                    .getValue(SubProtProviderExecution.PARAM_REQUESTED_RESOURCES);
        } else {
            // internal request
            state = MarketMakerConversationState.STATE_BUY_RESOURCES;
            requestedResource = new AllocationDescriptor();
        }
    }

    public boolean isFinished() {
        return state == MarketMakerConversationState.NO_ACTION;
    }

    public void action(ExtendedMessage in_message) {
        TextUi.println("ACTION at MarketMakerConversationState " + id);
        switch (state) {
            //@FIXME missing logic of follow-up of request execution
            case STATE_RESOURCE_REQUEST : {
                String offeredProvider = marketMaker
                        .getProviderAndAllocateResource(requestedResource);
                if (offeredProvider != null) {
                    Task task = (Task) in_message.getValue(SubProtProviderExecution.PARAM_TASK);
                    ExtendedMessage out_message = subProtProviderExecution
                            .createExecutionRequestMessage(offeredProvider,
                                                           requestedResource,
                                                           task);
                    marketMaker.send(out_message);
                    state = MarketMakerConversationState.STATE_RESOURCE_EXECUTE_RESPONSE;
                } else {
                    // drop conversation since request was rejected
                    ExtendedMessage out_message = subProtProviderExecution
                            .createExecutionRejectMessage(currentClient, id);
                    marketMaker.send(out_message);
                    state = MarketMakerConversationState.NO_ACTION;
                }
                break;
            }
            case STATE_RESOURCE_EXECUTE_RESPONSE: {
                ExtendedMessage out_message = null;
                // drop conversation in cases of terminal messages
                state = MarketMakerConversationState.NO_ACTION;
                if(in_message.getSubprotocolMessage().equals(
                        SubProtProviderExecution.MSG_EXECUTION_REJECT)){
                    out_message = subProtProviderExecution
                            .createExecutionRejectMessage(currentClient, id);
                }
                else if(in_message.getSubprotocolMessage().equals(
                        SubProtProviderExecution.MSG_EXECUTION_FINISHED)){
                    out_message = subProtProviderExecution
                            .createExecutionFinishedMessage(currentClient, id);
                }
                else if(in_message.getSubprotocolMessage().equals(
                        SubProtProviderExecution.MSG_EXECUTION_ACCEPT)){
                    out_message = subProtProviderExecution
                            .createExecutionFinishedMessage(currentClient, id);
                  // leave conversation in the pool
                  state = MarketMakerConversationState.STATE_RESOURCE_EXECUTE_RESPONSE;
                }
                marketMaker.send(out_message);
                break;
            }
            // Price request need not be followed by request. Drop conversation after answering
            case STATE_PRICE_REQUEST : {
                double price = marketMaker.getOffer(requestedResource);
                ExtendedMessage out_message = subProtProviderPriceQuery
                        .createPriceOfferMessage(currentClient, id, price);
                marketMaker.send(out_message);
                state = MarketMakerConversationState.NO_ACTION;
                break;
            }
            case STATE_BUY_RESOURCES : {
                currentInvestment = marketMaker.getInvestment();
                // keep trying
                if(currentInvestment <= 0.0)
                  break;
                ExtendedMessage out_message = subProtMatchmakerQuery
                        .createQueryMessage("MatchMaker_0", id, requestedResource);
                marketMaker.send(out_message);
                state = MarketMakerConversationState.STATE_BUY_MATCH;
                break;
            }
            case STATE_BUY_MATCH : {
                offeredProviders = (Collection<String>)in_message.getValue(
                                    SubProtMatchmakerQuery.PARAM_MULTIPLE_PROVIDERS_OFFERED);
                if(offeredProviders == null){
                  offeredProviders = new ArrayList<String>();
                  String offeredProvider = (String) in_message
                        .getValue(SubProtMatchmakerQuery.PARAM_OFFERED_PROVIDER);
                  offeredProviders.add(offeredProvider);
                }
                answers_count = offeredProviders.size();
                for(String offeredProvider : offeredProviders){
                  ExtendedMessage out_message = subProtProviderWholeSale
                          .createQueryMessage(offeredProvider, id, requestedResource);
                  marketMaker.send(out_message);
                }
                state = MarketMakerConversationState.STATE_BUY_ASK_PRICE;
                break;
            }
            case STATE_BUY_ASK_PRICE : {
                answers_count--;
                String provider = in_message.getSender();
                // if all providers rejected price request return to the beginning of the process
                if(in_message.getSubprotocolMessage().equals(
                            SubProtProviderWholeSale.MSG_OFFER_REJECT)){
                  offeredProviders.remove(provider);
                  if(offeredProviders.size() == 0)
                    state = MarketMakerConversationState.STATE_BUY_RESOURCES;
                  break;
                }
                if(providersOffers == null)
                  providersOffers = new ArrayList<Offer>();
                // assemble offers from all providers before proceeding
                Collection<AllocationDescriptor> priceOffers =
                      (Collection<AllocationDescriptor>)
                          in_message.getValue(SubProtProviderPriceQuery.PARAM_ASKING_PRICE);

                for(AllocationDescriptor item : priceOffers){
                  Offer offer = new Offer();
                  offer.allocationDescriptor = item;
                  offer.provider = provider;
                  providersOffers.add(offer);
                }
                if(answers_count == 0){
                  marketMaker.filterOffers(providersOffers);
                  // offers that are accepted make requests
                  answers_count = providersOffers.size();
                  if(answers_count == 0)
                    state = MarketMakerConversationState.STATE_BUY_RESOURCES;
                  else{
                    for(Offer offer : providersOffers){
                      ExtendedMessage out_message = subProtProviderWholeSale
                          .createSaleRequestMessage(offer.provider, id,
                                                    offer.allocationDescriptor);
                      marketMaker.send(out_message);
                    }
                    state = MarketMakerConversationState.STATE_BUY_ALLOCATE;
                  }
                }
                break;
            }
            case STATE_BUY_ALLOCATE : {
                answers_count--;
                String provider = in_message.getSender();
                // ignore rejected requests
                if(in_message.getSubprotocolMessage().equals(
                            SubProtProviderWholeSale.MSG_SALE_ACCEPT)){
                  AllocationDescriptor descriptor = (AllocationDescriptor)in_message
                              .getValue(SubProtProviderWholeSale.PARAM_REQUESTED_RESOURCES);
                  Offer offer = new Offer();
                  offer.allocationDescriptor = descriptor;
                  offer.provider = provider;
                  marketMaker.purchaseResource(offer);
                }
                // continue investment loop
                if(answers_count == 0)
                  state = MarketMakerConversationState.STATE_BUY_RESOURCES;
                break;
            }
        }
    }
}
