package yaes.application.uavaltclient.agents;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Set;

import yaes.application.uavaltclient.missionmgr.InitialAssignment;
import yaes.application.uavaltclient.missionmgr.Mission;
import yaes.application.uavaltclient.missionmgr.Objective;
import yaes.application.uavaltclient.naming.ExtendedName;
import yaes.application.uavaltclient.ui.UIServer;
import yaes.application.uavaltclient.message.MsgType;
import yaes.application.uavaltclient.message.MessageLibrary;

import yaes.framework.agent.AbstractCommunicatingAgent;
import yaes.framework.agent.GenericMessage;
import yaes.framework.world.World;

/**
 * HQ, the headquarters agent, acts on behalf of the user, issuing commands from and returning 
 * results to the user.  Also sends out inital assignments to UAVs.
 * 
 * @author dejongp
 */
public class HQ extends AbstractCommunicatingAgent
{
  private static final String HQ_NAME = "headquarters";
	
	private UIServer ui = null;
	
	private HashMap directoryMap = new HashMap(); // maps extended-name/squadron --> HashMap (of uav name-->uav name)
	private HashMap namingMap = new HashMap(); // maps extendedName.toString() --> ExtendedName obj
	private HashMap handleMap = new HashMap(); //maps handle --> squadron 
  
  // msg communication state info
	private Mission newMission = null;
	private boolean missionMergePending = false; 
	private boolean squadronMergePending = false;
	private String[] squadronsAwaitingMerge = new String[2];
	private Mission[] squadronMergeMissions = new Mission[2];
	private boolean squadronSplitPending = false;
	

  /**
   * Constructor.  HQ must be constructed AFTER UAVs, so that initial assignments are received. 
   */
  public HQ(World world)
  {
  	super(HQ_NAME, world);
  }
  
  /**
   * Initialize the first mission and send initial assignment to uavs.
   * 
   * @param initialAssignment an instance of the InitialAssignment class, containing startup values
   */
  public void initialize(InitialAssignment initialAssignment)
  {
		for (int i=0; i<initialAssignment.getExtendedNames().size(); i++)
		{
			ExtendedName ename = (ExtendedName) initialAssignment.getExtendedNames().get(i);
			namingMap.put(ename.toString(), ename);
			handleMap.put("s" + (i+1), ename.toString());
			directoryMap.put(ename.toString(), new HashMap());
		}
		
		// send initial assignments to UAVs
		HashMap squadronUAVs = initialAssignment.getSquadronUAVs();
		Iterator iter = squadronUAVs.keySet().iterator();
		while (iter.hasNext() )
		{
			String uav = (String) iter.next();
			String squadName = (String) squadronUAVs.get(uav);
			Mission mission = new Mission((String) initialAssignment.getMissions().get(squadName));

			MessageLibrary.sendNewSquadronMsg(HQ_NAME, uav, squadName, initialAssignment.getMembers());
			MessageLibrary.sendNewMissionMsg(HQ_NAME, uav, (Mission) mission.clone(), new Boolean(true));
			
			((HashMap) directoryMap.get(squadName)).put(uav, uav);
			
      // old assign-leader, assign-formation was here
		}		
  }

  /**
   * Main control loop for the agent.  Req'd by superclass.
   * 
   * @see yaes.framework.agent.IAgent#action(yaes.framework.world.World)   
   */
  public void action()
  {
  	if (ui == null)
  	{
  		ui = new UIServer(this);
  		Thread t = new Thread(ui);
  		t.start();
  	}
  	
		processMessages();
  }

	/**
	 * Sends a command message for the user.
	 * 
	 * @param msg String command sent from user.
	 */
	public void command(String cmd)
	{
		if (cmd.startsWith("cmd-new-mission "))
			processNewMissionCommand(cmd);
		else if (cmd.startsWith("cmd-merge-mission "))
			processMergeMissionCommand(cmd);
		else if (cmd.startsWith("cmd-merge "))
			processMergeCommand(cmd);
		else if (cmd.startsWith("cmd-split "))
			processSplitCommand(cmd);
		else if (cmd.startsWith("cmd-get-current-objective "))
			processGetCurrentObjectiveCommand(cmd);
		else if (cmd.startsWith("cmd-get-objectives "))
			processGetObjectivesCommand(cmd);			
		else if (cmd.startsWith("cmd-who")) 
			who();
	}
	  
	/**
	 * Process incoming messages.
	 */
	private void processMessages()
	{
		for (int msgCnt = 0; msgCnt < receivedMessages.size(); msgCnt++)
		{
			GenericMessage msg = (GenericMessage) receivedMessages.remove(0);
			String msgType = (String) msg.getValue("type");

      // uav has sent an objective intended for display to the user
      if (MsgType.MSG_CURRENT_OBJECTIVE.equals(msgType))
      {
        Objective obj = (Objective) msg.getValue("objective");
        ui.write((String) obj.toString());
      }
      
      // uav has sent a list of all objectives intended for display to the user
      else if (MsgType.MSG_OBJECTIVES.equals(msgType))
        ui.write((String) msg.getValue("objectivesDescription"));
      
      // uav has sent a mission, in response to user merge mission, merge squad, or split squad cmd
			else if (MsgType.MSG_MISSION.equals(msgType))
			{
				String squadName = findSquadron((String) msg.getValue("sender")); 
				Mission mission = (Mission) msg.getValue("mission");
				int objectiveIndex = ((Integer) msg.getValue("objectiveIndex")).intValue();			
				
				// merge mission sequence is underway
				if (missionMergePending)
				{
					mergeMissions(newMission, mission, objectiveIndex, squadName);
					newMission = null;	
					missionMergePending = false;
				}

        // merge squadron sequence is underway
				else if (squadronMergePending)
				{
					if (squadronMergeMissions[0] == null)
					  squadronMergeMissions[0] = mission;
					else
					{
						squadronMergeMissions[1] = mission;
						mergeSquadrons();
						
						squadronMergePending = false;	
						squadronsAwaitingMerge = null;
						squadronMergeMissions = null;
						who();						
					}			
				}
				
        // split squadron sequence is underway
				else if (squadronSplitPending)
				{
					String oldSquadron = findSquadron((String) msg.getValue("sender"));
					splitSquadrons(mission, oldSquadron);			
					squadronSplitPending = false;
					who();
				}		
			}
		  
			else
				System.err.println("MESSAGE NOT UNDERSTOOD: " + msgType);
		}
 
	}

	/**
	 * Merge two missions, sending the result to all members of a specified squadron.
	 * 
	 * @param newMission mission to merge into original mission
	 * @param mission original mission to be merged with newMission specified by user
	 * @param objectiveIndex current objective index inside original mission
	 * @param squadName name of squadron which will receive the newly merged mission 
	 */
  private void mergeMissions(Mission newMission, Mission mission, int objectiveIndex, String squadName)
  {
  	System.out.println("HQ: MERGING MISSIONS FOR SQUAD " + squadName);
		mission.mergeMission(newMission, objectiveIndex);
				
		HashMap uavs = (HashMap) directoryMap.get(squadName);			
		Iterator uavIter = uavs.values().iterator();
			
		while (uavIter.hasNext() )
		{
			String receiver = (String) uavIter.next();
			MessageLibrary.sendNewMissionMsg(HQ_NAME, receiver, (Mission) mission.clone(), new Boolean(false));
		}
  }
  
  /**
   * Merge two squadrons into one.  Provide new mission and squadron assignments to members.
   */
  private void mergeSquadrons()
	{
		ExtendedName newSquadron = 
			ExtendedName.merge((ExtendedName) namingMap.get(squadronsAwaitingMerge[0]),
												 (ExtendedName) namingMap.get(squadronsAwaitingMerge[1]));
   
		namingMap.put(newSquadron.toString(), newSquadron);
			
		// update Directory structures to reflect new UAV squadron assignments
		ArrayList squad1 = getMemberNames(squadronsAwaitingMerge[0]);
		ArrayList squad2 = getMemberNames(squadronsAwaitingMerge[1]);
			
		for (int i=0; i<squad1.size(); i++)
			moveUAV((String) squad1.get(i), newSquadron.toString(), squadronsAwaitingMerge[0]);					
			
		for (int i=0; i<squad2.size(); i++)
			moveUAV((String) squad2.get(i), newSquadron.toString(), squadronsAwaitingMerge[1]);				
					
		mergeMissions(squadronMergeMissions[0], squadronMergeMissions[1], 0, newSquadron.toString());
					
		ArrayList membersList = getMemberNames(newSquadron.toString());
		String[] members = new String[membersList.size()];
		for (int i=0; i<membersList.size(); i++)
			members[i] = (String) membersList.get(i);
					
		HashMap uavs = (HashMap) directoryMap.get(newSquadron.toString());			
		Iterator uavIter = uavs.values().iterator();
								
		while (uavIter.hasNext() )
		{
			String receiver = (String) uavIter.next();
			MessageLibrary.sendNewSquadronMsg(HQ_NAME, receiver, newSquadron.toString(), (String[]) members.clone());
		}		
					
		squadronMergePending = false;				
	}
  
	/**
	 * Split one squadron into two.  Provide new mission and squadron assignments to members of both
	 * squadrons.
	 * 
	 * @param mission orignal mission to split
	 * @param oldSquadron name of the old squadron, before split
	 */
	private void splitSquadrons(Mission mission, String oldSquadron)
	{
		Mission mission2 = mission.splitMission();
		ArrayList squad = getMemberNames(oldSquadron);
		ArrayList newExtendedNames = ((ExtendedName) namingMap.get(oldSquadron)).split();
		ExtendedName newSquadron1 = (ExtendedName) newExtendedNames.get(0);
		ExtendedName newSquadron2 = (ExtendedName) newExtendedNames.get(1);
			
		namingMap.put(newSquadron1.toString(), newSquadron1);
		namingMap.put(newSquadron2.toString(), newSquadron2);
			
		for (int i=0; i<squad.size(); i++)
		{
			if (i < squad.size() / 2)				
				moveUAV((String) squad.get(i), newSquadron1.toString(), oldSquadron);
			else
				moveUAV((String) squad.get(i), newSquadron2.toString(), oldSquadron);
		}
  
		ArrayList newsquad1 = getMemberNames(newSquadron1.toString());
		ArrayList newsquad2 = getMemberNames(newSquadron2.toString());
					
		String[] newSquad1members = new String[newsquad1.size()];
		for (int i=0; i<newsquad1.size(); i++)
			newSquad1members[i] = (String) newsquad1.get(i);					
					
		String[] newSquad2members = new String[newsquad2.size()];
		for (int i=0; i<newsquad2.size(); i++)
			newSquad2members[i] = (String) newsquad2.get(i);		
											
		assignNewMission(mission, newSquadron1.toString());
		assignNewMission(mission2, newSquadron2.toString());
		assignNewSquadron(newSquadron1.toString(), newSquad1members);
		assignNewSquadron(newSquadron2.toString(), newSquad2members);
	}
	
	/**
	 * Move a UAV from one squadron to another. 
	 * 
	 * @param uavName
	 * @param destSquadname
	 * @param origSquadname may be null if no initial squadname.
	 */
	private void moveUAV(String uavName, String destSquadname, String origSquadname)
	{
		// remove from current group, if any
		if (origSquadname != null)
		{
			HashMap groupMap = (HashMap) directoryMap.get(origSquadname);
			if (groupMap != null)
			{
				groupMap.remove(uavName);
		  	
				if (groupMap.size() == 0)
					directoryMap.remove(origSquadname);
			}
		}
		
		// add aid to new group
		HashMap groupMap = (HashMap) directoryMap.get(destSquadname);
		if (groupMap == null)
		{
			groupMap = new HashMap();
			directoryMap.put(destSquadname, groupMap);
		}

		groupMap.put(uavName, uavName);	
	}
  
  /**
   * Send a new mission to all UAVs in the squadron, unless the squadron name is invalid.
   * 
   * @param cmd String command sent from user.
   */
	private void processNewMissionCommand(String cmd)
	{
		String[] tokens = cmd.split(" ");
		String missionFile = tokens[1];
		String squadName = (String) handleMap.get(tokens[2]);		
		Mission mission = new Mission(missionFile);
		assignNewMission(mission, squadName);
	}
	
	/**
	 * Sends a copy of the mission to every uav in a squadron. 
	 * 
	 * @param mission
	 * @param squadron
	 */
	private void assignNewMission(Mission mission, String squadName)
	{
		if (directoryMap.containsKey(squadName))
		{
			HashMap uavs = (HashMap) directoryMap.get(squadName);			
			Iterator iter = uavs.values().iterator();
			
			while (iter.hasNext() )
			{
				String receiver = (String) iter.next();
				MessageLibrary.sendNewMissionMsg(HQ_NAME, receiver, (Mission) mission.clone(), new Boolean(false));
			}
		}
		else
		{
			System.err.println("ERROR: Squadname " + squadName + " not valid.");
		}		
	}
	
	/**
	 * Sends new squadron assignments to every uav in a squadron
	 * 
	 * @param squadName
	 */
	private void assignNewSquadron(String squadName, String[] members)
	{
		if (directoryMap.containsKey(squadName))
		{
			HashMap uavs = (HashMap) directoryMap.get(squadName);			
			Iterator iter = uavs.values().iterator();
			
			while (iter.hasNext() )
			{
				String receiver = (String) iter.next();
				MessageLibrary.sendNewSquadronMsg(HQ_NAME, receiver, squadName, (String[]) members.clone() );
			}
		}
		else
		{
			System.err.println("ERROR: Squadname " + squadName + " not valid.");
		}	
	}

	/**
	 * Request a mission to be sent so it can be merged.
	 * 
	 * @param cmd String command sent from user.
	 */
	private void processMergeMissionCommand(String cmd)
	{
		String[] tokens = cmd.split(" ");
		String missionFile = tokens[1];
		String squadName = (String) handleMap.get(tokens[2]);		
		
		if (directoryMap.containsKey(squadName))
		{
			newMission = new Mission(missionFile);
			HashMap uavs = (HashMap) directoryMap.get(squadName);			
			Iterator iter = uavs.values().iterator();
			missionMergePending = true;
			
			// send mission request to a single uav
			if (iter.hasNext() )
			{
				String receiver = (String) iter.next();
				MessageLibrary.sendRequestMissionMsg(HQ_NAME, receiver);
			}
		}
		else
		{
			ui.write("ERROR: Squadname not valid.");
		}
	}
	
	/**
	 * Initiate sequence to merge two squadrons.
	 * 
	 * @param msg String command sent from user.
	 */
	private void processMergeCommand(String cmd)
	{
		String[] tokens = cmd.split(" ");
		String squad1 = (String) handleMap.get(tokens[1]);
		String squad2 = (String) handleMap.get(tokens[2]);

		// obtain members for squadrons, preferably squadron leaders
		String squad1Recipient = obtainMember(squad1);
		String squad2Recipient = obtainMember(squad2);	
		
		squadronMergePending = true;
		squadronMergeMissions = new Mission[2];
		squadronsAwaitingMerge = new String[] {squad1, squad2};
				
		MessageLibrary.sendRequestMissionMsg(HQ_NAME, squad1Recipient);
		MessageLibrary.sendRequestMissionMsg(HQ_NAME, squad2Recipient);
	}		
	
	/**
	 * Initiate sequence to split a squadron in two.
	 * 
	 * @param cmd
	 */
	private void processSplitCommand(String cmd)
	{
		String[] tokens = cmd.split(" ");
		String squad = (String) handleMap.get(tokens[1]);

		// obtain members for squadrons, preferably squadron leaders
		String recipient = obtainMember(squad);
		squadronSplitPending = true;
		MessageLibrary.sendRequestMissionMsg(HQ_NAME, recipient);
	}
  
  /**
   * Initiate sequence to return current objective to user. 
   * 
   * @param cmd
   */
	private void processGetCurrentObjectiveCommand(String cmd)
	{
		String[] tokens = cmd.split(" ");
		String squad = (String) handleMap.get(tokens[1]);
		
		String recipient = obtainMember(squad);
		MessageLibrary.sendGetCurrentObjectiveMsg(HQ_NAME, recipient);
	}
	
	/**
	 * Initiate sequence to display all objectives for user. 
	 * 
	 * @param cmd
	 */
	private void processGetObjectivesCommand(String cmd)
	{
		String[] tokens = cmd.split(" ");
		String squad = (String) handleMap.get(tokens[1]);
		
		String recipient = obtainMember(squad);
		MessageLibrary.sendGetObjectivesMsg(HQ_NAME, recipient);
	}
		
	/**
	 * Obtains name of one member of the specified squad.
	 *  
	 * @param squadName
	 * @return a member of squad, or null if none are found
	 */
	private String obtainMember(String squadName)
	{
		HashMap uavs = (HashMap) directoryMap.get(squadName);			
		Iterator iter = uavs.values().iterator();
		
		// send mission request to a single uav
		if (iter.hasNext() )
		{
			String member = (String) iter.next();
			return member;
		}

    return null;
	}

  /**
   * Inform user names and contents of all squadrons.  Assigns handles to squadron extended names.
   */
  private void who()
  {
  	String response = "";
  	
  	Set keys = directoryMap.keySet();
  	Iterator iter = keys.iterator();
  	
  	handleMap = new HashMap();
  	int handleNumber = 0;
  	
  	while (iter.hasNext())
  	{
  		String extendedName = (String) iter.next();
			response += extendedName + "  HANDLE: s" + (++handleNumber) + "\r\n";
			handleMap.put("s" + handleNumber, extendedName);
  		HashMap groupMap = (HashMap) directoryMap.get(extendedName);
  	  Set groupKeys = groupMap.keySet();
  	  Iterator groupIter = groupKeys.iterator();
  	  
  	  while (groupIter.hasNext() )
  	  {
  	  	String aidstring = (String) groupIter.next();
  	  	response += "  --> " + aidstring + "\r\n";
  	  }
  	  response += "\r\n";
  	}
  	
  	ui.write(response);
  }
		
	/**
	 * Return squadron name for the specified UAV.
	 * 
	 * @param name name for the uav to find squadron for
	 * @return squadron to which the specified uav belongs, or null if not found
	 */
	private String findSquadron(String name)
	{
		Set keys = directoryMap.keySet();
		Iterator iter = keys.iterator();
		while (iter.hasNext())
		{
			String squadron = (String) iter.next();
			HashMap groupMap = (HashMap) directoryMap.get(squadron);
			if (groupMap.containsKey(name))
				return squadron;
		}
		
		System.err.println("findSquadron(String name) could not find squadron for " + name);
		return null;	
	}

  /**
   * Returns an arraylist containing all member aid strings for a given squadron
   * 
   * @param squadname
   * @return
   */
  private ArrayList getMemberNames(String squadname)
  {
  	ArrayList names = new ArrayList();
  	
    HashMap groupMap = (HashMap) directoryMap.get(squadname);
    Iterator iter = groupMap.keySet().iterator();
    while (iter.hasNext())
    {
      String aidString = (String) iter.next();
      names.add(aidString);
    }
    return names;
  }
}
