package yaes.application.uavaltclient.agents;

import java.util.ArrayList;
import java.util.Calendar;

import yaes.application.uavaltclient.missionmgr.Mission;
import yaes.application.uavaltclient.missionmgr.MissionManager;
import yaes.application.uavaltclient.missionmgr.Objective;
import yaes.application.uavaltclient.nav.NavUtil;
import yaes.application.uavaltclient.nav.Pos;
import yaes.application.uavaltclient.simulation.SimException;
import yaes.application.uavaltclient.simulation.SimProxy;
import yaes.application.uavaltclient.message.MsgType;
import yaes.application.uavaltclient.message.MessageLibrary;

import yaes.framework.agent.AbstractCommunicatingAgent;
import yaes.framework.agent.GenericMessage;
import yaes.framework.world.World;
 

/**
 * Abstract superclass for UAV strategies.
 * 
 * @author dejongp
 */
public abstract class UAV extends AbstractCommunicatingAgent
{	
	protected String type; //uav type
  protected SimProxy sim;
  
	protected MissionManager mgr = null;
	
  // name of this squadron, used to multicast messages to other UAVs in squadron
	private String squadronName = null;
	
	// names of all members of this squadron
	protected String[] squadronMembers;
	
	protected String name = null; // this UAV agent's name

	private ArrayList ldrProposals; //storage container for leader proposals during negotiation
  private boolean ldrNegotiationFlag = false;
  private int numLdrProposalsExpected = 0;
	
	protected boolean bombNegotiationFlag = false;
	
	private long startTime;

	private double firepower;
	private double visibility;
	private double maxVelocity;
	private double stealth;

	
	/**
	 * Obtain the maximum heading change per tick for this UAV.
	 * 
	 * @return max heading change per tick (degrees)
	 */
	public abstract double getMaxHdgDelta();
	
	/**
	 * Obtain the maximum altitude change per tick for this UAV.
	 * 
	 * @return max heading change per tick (degrees)
	 */
	public abstract double getMaxAltDelta();	
	
	
  /**
	 * Handling of subclass-specific messages.
	 *
	 * @param msg the incoming ACL message
	 * @param content the content of the incoming ACLMessage
	 * @param msgType the type of the message (one of those contained in MsgType class)
	 */	
	protected abstract void processMessage(GenericMessage msg, String msgType);
	
  /**
   * 
   * @param name name of the uav, used for messaging
   */
  public UAV(String name, World world, double firepower, double visibility, double maxVelocity, double stealth)
  {  	
    super(name, world);
    
    this.name = name;
    this.firepower = firepower;
    this.visibility = visibility;
    this.maxVelocity = maxVelocity;
    this.stealth = stealth;

    try
    {
      sim = new SimProxy();
    }
    catch(SimException e) 
    {
    	System.err.println(e);
    	e.printStackTrace();
    }

    startTime = Calendar.getInstance().getTimeInMillis();	
  }
	

  /**
   * Main control loop for the UAV. Use this algorithm:
   *   - Process messages / get Positions of other aircraft
   *   - Notify other aircraft in squadron of my position
   *   - Determine next move based on aircraft state and mission
   *   - Send movement cmd to Simulation (via SimProxy)
   * 
   * @param world YAES context
   * @see yaes.framework.agent.IAgent#action(yaes.framework.world.World)
   */
  public void action()
  {
	  // Process messages / get Positions of other aircraft
	  processMessages();
	 
    if (mgr != null)
    {
      try
      {
        Pos curPos = getPosFromSim();
        mgr.setCurrentPosition(curPos);
        
        // send out current position to other UAVs
				for (int i=0; i<squadronMembers.length; i++)
					if (! name.equals(squadronMembers[i]))
						MessageLibrary.sendPosMsg(name, squadronMembers[i], curPos);
        
        ///System.out.println(curPos.getString());
        sim.rangeFinder(mgr.getMap(), 4, Math.round((float) curPos.getY()));
        // examine surrounding area
        mgr.computeNextMove(); // will compute next move based on position and mission
        sim.move(mgr.getNextVel(), mgr.getNextHgt(), Math.toRadians(mgr.getNextHdg()));
      }
      catch (SimException e)
      {
      	System.err.println("SIM EXCEPTION! " + e.getMessage());
      	e.printStackTrace();
      }
    }
  }

  /**
   * Process incoming messages.
   */
	private void processMessages()
	{
		for (int msgCnt = 0; msgCnt < receivedMessages.size(); msgCnt++)
		{
			GenericMessage msg = (GenericMessage) receivedMessages.remove(0);
		  String msgType = (String) msg.getValue("type");

      // update the map with another aircraft's position
			if (MsgType.MSG_POS.equals(msgType))
			  mgr.getMap().updateAircraft((String) msg.getValue("sender"), (Pos) msg.getValue("pos"));

      // replace the current mission with a new mission
			else if (MsgType.MSG_NEW_MISSION.equals(msgType))
			{
				mgr = new MissionManager(this, (Mission) msg.getValue("mission"));
			}				

      // send a clone of the current mission to the requester
			else if (MsgType.MSG_REQUEST_MISSION.equals(msgType))
				MessageLibrary.sendMissionMsg(name, 
				                              (String) msg.getValue("sender"), 
				                              (Mission) mgr.getMission().clone(),
				                              new Integer(mgr.getCurrentObjectiveIndex()));
				
			// New squadron assignment.  Initiate a leader negotiation if one has not already been initiated.  
			else if (MsgType.MSG_NEW_SQUADRON.equals(msgType))
			{
        squadronName = (String) msg.getValue("squadron");
        squadronMembers = (String[]) msg.getValue("members");
        
        // this should trigger a leader negotiation, necessary when subsequent new squadrons are formed
        if (mgr != null)
        {  
					mgr.setLeader(null);
					
        	if (squadronMembers.length > 1)
	          mgr.setCurrentMode(MissionManager.MODE_FORMATION);
        	else
					  mgr.setCurrentMode(MissionManager.MODE_NAV);
        }
        
        if (this instanceof Scout)
          ((Scout) this).resetTargetsFound();
			}
		  
			// send a clone of the current objective to the requester
			else if (MsgType.MSG_GET_CURRENT_OBJECTIVE.equals(msgType))
				MessageLibrary.sendCurrentObjectiveMsg(name, 
                                               (String) msg.getValue("sender"), 
                                               (Objective) mgr.getCurrentObjective().clone());
		  
			// send a String description of mission objectives to the requester
			else if (MsgType.MSG_GET_OBJECTIVES.equals(msgType))
			  MessageLibrary.sendObjectivesMsg(name, 
																				(String) msg.getValue("sender"), 
																				(String) mgr.getMission().getObjectiveList(0));		  
		  
			// respond to a leader negotiation call-for-proposal
			else if (MsgType.MSG_LEADER_NEEDED.equals(msgType))
			{
				ldrNegotiationFlag = true;
				MessageLibrary.sendProposeLeadMsg(name, 
				                                  (String) msg.getValue("sender"), 
				                                  (Pos) mgr.getCurrentPosition(), 
				                                  type);
			}
		  
			// this uav has been assigned leader as a result of a leader negotiation
			else if (MsgType.MSG_ASSIGN_LEADER.equals(msgType))
			  leaderAssignment();
			
			// this uav has been assigned formation flight as a result of a leader negotiation
			else if (MsgType.MSG_ASSIGN_FORMATION.equals(msgType))
				formationAssignment((String) msg.getValue("leader"), 
				                    ((Integer) msg.getValue("row")).intValue(), 
				                    (String) msg.getValue("side"));
		  
			// this uav has been assigned wait loiter as a result of a bomber negotiation
			else if (MsgType.MSG_ASSIGN_WAIT.equals(msgType))
			{
				mgr.setWaitPosition(mgr.getCurrentPosition());
				mgr.setCurrentMode(MissionManager.MODE_WAIT);
			}				  

			// another uav has responded to this uav's leader negotiation CFP
			else if (MsgType.MSG_PROPOSE_LEAD.equals(msgType))
			{
				ldrProposals.add(msg);
				numLdrProposalsExpected --;
				if (numLdrProposalsExpected == 0)
					selectLeader();
			}
			
			// this uav is instructed to resume mission after a bomb run is carried out by another aircraft
			else if (MsgType.MSG_RESUME_MISSION.equals(msgType))
			{
				bombNegotiationFlag = false;
				mgr.resumeMission();
			}

			// leader has transitioned to next objective, notifies squadron of current objective index
			else if (MsgType.MSG_OBJECTIVE_NUM.equals(msgType))
			{
				mgr.setCurrentObjectiveIndex(((Integer) msg.getValue("objectiveIndex")).intValue());
			}
			
			// message cannot be handled by UAV;  let subclass process it
			else
				processMessage(msg, msgType);
		}
	}

	/**
	 * Initiate a leader negotiation.
	 */
	public void initiateLeaderNegotiation()
	{
		// don't initiate ldr negotiation if one is already underway
		if (ldrNegotiationFlag || isLdrNeededMsgInQueue() )
		{
			//System.out.println("uav " + name + " of squad " + this.squadronName + " ignores leader neg: ");
			//if (ldrNegotiationFlag) System.out.println(" ldrNegFlag is true");
			//if (isLdrNeededMsgInQueue() ) System.out.println(" isLdrNeededMsgInQueue() is true");
		
		  return;
		}
		
		System.out.println("uav " + name + " of squad " + this.squadronName + " initiates Ldr Neg!!: ");
		//if (ldrNegotiationFlag) System.out.println(" ldrNegFlag is true");
		//if (isLdrNeededMsgInQueue() ) System.out.println(" isLdrNeededMsgInQueue() is true");
					
		ldrNegotiationFlag = true;
		numLdrProposalsExpected = squadronMembers.length - 1;
		ldrProposals = new ArrayList();

		for (int i=0; i<squadronMembers.length; i++)
			if (! name.equals(squadronMembers[i]))
				MessageLibrary.sendLeaderNeededMsg(name, squadronMembers[i]);
	}
	
	/**
	 * Peek in the queue, look for MSG_LEADER_NEEDED.  This allows uav to avoid leader negotiation
	 * when one has already been initiated.  Necessary since otherwise all UAVs in squadron detect
	 * missing leader simultaneously. 
	 * 
	 * @return true if msg queue contains MSG_LEADER_NEEDED, false otherwise
	 */
	private boolean isLdrNeededMsgInQueue()
	{
		for (int msgCnt = 0; msgCnt < receivedMessages.size(); msgCnt++)
		{
			GenericMessage msg = (GenericMessage) receivedMessages.get(msgCnt);
			if (((String) msg.getValue("type")).equals(MsgType.MSG_LEADER_NEEDED))
			{
				//System.out.println("** " + getName() + ": FOUND MSG_LEADER_NEEDED IN QUEUE");
				return true;
			}
		}
		//System.out.println("-- " + getName() + ":  DID NOT FIND MSG_LEADER_NEEDED IN QUEUE");
		return false;
	}
	
	/**
	 * Select a leader from the candidates as part of a leader negotiation.
	 * Precondition: all proposals have been received.
	 */
	private void selectLeader()
	{
		String bestType = type;
		String bestCandidate = name;
		Pos bestPos = mgr.getCurrentPosition();
		Pos objPos = mgr.getCurrentObjective().getPos1();
		
		for (int i=0; i<ldrProposals.size(); i++)
		{
			GenericMessage proposalMsg = (GenericMessage) ldrProposals.get(i);
			Pos proposedPos = (Pos) proposalMsg.getValue("pos");
			String proposedType = (String) proposalMsg.getValue("uavType");
		  		  	
			if ((proposedType == UAVType.UAV_TYPE_NORMAL &&
					(!(bestType == UAVType.UAV_TYPE_NORMAL)))
			 || (proposedType == UAVType.UAV_TYPE_SCOUT &&
					 bestType == UAVType.UAV_TYPE_BOMBER)
			 || (NavUtil.distance(proposedPos.getX(), proposedPos.getZ(), objPos.getX(), objPos.getZ()) < 
					 NavUtil.distance(bestPos.getX(), bestPos.getZ(), objPos.getX(), objPos.getZ())))
			 {
				 bestCandidate = (String) proposalMsg.getValue("sender");
				 bestPos = proposedPos;
				 bestType = proposedType;
			 } 
		}
		
		int row = 0;
		String side = "right";
		
		for (int i=0; i<squadronMembers.length; i++)
		{
			// send formation slot assignment
		  if (! squadronMembers[i].equals(bestCandidate))
		  {
		  	if (side.equals("right"))
		  	{
		  		side = "left";
		  		row ++;
		  	}
		  	else
		  		side = "right";
		  		
		  	// uav other than initiator
		  	if (! squadronMembers[i].equals(name))
					MessageLibrary.sendAssignFormationMsg(name, squadronMembers[i], bestCandidate, new Integer(row), side);
			  // initiator assigns itself formation flight 
		  	else
		  	  formationAssignment(bestCandidate, row, side);
		  }

		  // send leader assignment
      else
      {
				// uav other than initiator is assigned leader
				if (! squadronMembers[i].equals(name))
					MessageLibrary.sendAssignLeaderMsg(name, squadronMembers[i]);
				// initiator assigns itself leader
				else
					leaderAssignment();
      }
		}
		
		// exit leader negotiation state
		ldrNegotiationFlag = false;
		numLdrProposalsExpected = 0;
		ldrProposals = null;
	}
	
	/**
	 * Prepares internal state for formation flight.  Result of a leader negotiation.
	 * 
	 * @param leader name of the squadron leader
	 * @param rank number of rows behind leader in formation
	 * @param side either "left" or "right" of the leader
	 */
	private void formationAssignment(String leader, int rank, String side)
	{
		ldrNegotiationFlag = false; // turns off leader negotiation for participants
		mgr.setLeader(leader);
		mgr.setFormationRank(rank);
		mgr.setFormationSide(side);
		mgr.setCurrentMode(MissionManager.MODE_FORMATION);
	}		
	
	/**
	 * Prepares internal state for leadership.  Result of a leader negotiation.
	 */
	private void leaderAssignment()
	{
		ldrNegotiationFlag = false; // turns off leader negotiation for participants
		mgr.setLeader(name);
		mgr.setCurrentMode(MissionManager.MODE_NAV);
		mgr.reloadCurrentObjective();	
	}

  /**
   * This UAV has been assigned a follower.
   * 
   * @param content
   */
	/*
	private void processAssignFollowMsg(String content)
	{
		System.out.println(getTheBondAgent().getAID().getLocalName() + " ASSIGNED FOLLOWER");
		
		String[] tokens = content.split(" ");
		mgr.setLeader(tokens[1]);
		mgr.setCurrentMode(MissionManager.MODE_FOLLOW);
		
		ACLMessage msg = new ACLMessage(ACLMessage.INFORM);
		msg.setContent("assignment follow ack");
		msg.addReceiver(EnhancedDirectory.getDirectoryAID());
		getTheBondAgent().send(msg);
		
		if (MSG_RECORD_FLAG)
			incrementMsg("assignment");		
	}
	*/

			
	/**
	 * Obtain number of seconds this UAV has been in existance.
	 * 
	 * @return number of seconds since instantiation.
	 */
	public long elapsedTime()
	{
		return (Calendar.getInstance().getTimeInMillis() - startTime) / 1000;
	}

  /**
   * @return this UAV's type
   */
  public String getType()
  {
    return type;
  }
	
	/**
	 * Obtain this UAV's current position from the simulation.
	 * 
	 * @return current position
	 * @throws SimException
	 */
	private Pos getPosFromSim() throws SimException
	{
		Pos curPos = sim.position();
		curPos.setHdg(sim.angle());
		curPos.setV(sim.velocity());
		curPos.setAlt(sim.height());
		return curPos;
	}
	
  /**
   * @return
   */
  public String[] getSquadronMembers()
  {
    return squadronMembers;
  }

  /**
   * @return
   */
  public String getName()
  {
    return name;
  }

  /**
   * @return
   */
  public double getFirepower()
  {
    return firepower;
  }

  /**
   * @return
   */
  public double getMaxVelocity()
  {
    return maxVelocity;
  }

  /**
   * Cruise velocity: 3/4 of maximum velocity.
   * 
   * @return
   */
  public double getCruiseVelocity()
  {
  	return maxVelocity * 0.75;
  }

  /**
   * @return
   */
  public double getStealth()
  {
    return stealth;
  }

  /**
   * @return
   */
  public double getVisibility()
  {
    return visibility;
  }

}

