

package yaes.application.uavaltclient.coalitionformation;



import java.math.BigInteger;

import java.util.ArrayList;

import java.util.Collection;

import java.util.HashMap;

import java.util.Iterator;

import java.util.List;

import java.util.Map;

import java.util.Set;



import yaes.application.uavaltclient.coalitionformation.timing.*;



/**

 * Abstract superclass for all coalition formation, regardless of formation algorithm.

 * 

 * @author dejongp

 */

public abstract class CoalitionFormation

{

	protected static final int MAX_QUALITY_TIMING_CNT = 200; // max number of quality data points which can be collected

	

	protected List candidates = new ArrayList(); //UAVRepresentation[]

	protected List missions = new ArrayList(); //MissionRepresentation[]

	protected List coalitions = new ArrayList(); //Coalition[] .  Added by concrete subclass.

	

	// timestamp values for entire run (all missions), set by concrete subclass.

	protected Timing totalTime = Timing.getInstance();

	protected Map missionFirstSolutionTimes = new HashMap(); // name=missionName, key=Timing. Each Timing instance holds time for one mission computation

	

	// for measuring quality increase (i.e, lowering cost) versus time

	protected int[] qualityCost = new int[MAX_QUALITY_TIMING_CNT]; // Integer[] of cost values 

	protected double[] qualityTime = new double[MAX_QUALITY_TIMING_CNT]; // Double[] of timestamp values

	protected Timing[] qualityTimingObjects = new Timing[MAX_QUALITY_TIMING_CNT];

	protected boolean qualityTimingFlag = false;

	protected boolean breakAfterFirstSolution = false;

	

	/**

	 * Protected constructor

	 */

	protected CoalitionFormation()

	{

	}

	

	/**

	 * Instruct coalition formation to cease after first discovered solution.

	 *

	 */

	public void breakAfterFirstSolution()

	{

	  this.breakAfterFirstSolution = true;	

	}

	

	/**

	 * Enable quality timing, for certain experiments.

	 */

  public void enableQualityTiming()

  {

    qualityTimingFlag = true;

    

		for (int i=0; i<MAX_QUALITY_TIMING_CNT; i++)

			qualityTimingObjects[i] = Timing.getInstance();    

  }

	

	/**

	 * Start all quality timing list timers.

	 */

	protected void startQualityTimers()

	{

		for (int i=0; i<MAX_QUALITY_TIMING_CNT; i++)

		  qualityTimingObjects[i].start();

	}

	

	/**

	 * Return the average first solution time for all the missions.  Must be called after formCoalitions() has 

	 * successfully executed!

	 * 

	 * @return

	 */

	public double getAvgFirstSolutionTime()

	{

		double sum = 0.0;

		

	  for (int i=0; i<coalitions.size(); i++)

	  {

	  	Coalition coalition = (Coalition) coalitions.get(i);

	  	MissionRepresentation mRep = coalition.getMissionRep();	

	  	Timing t = (Timing) missionFirstSolutionTimes.get(mRep.getName());

	  	try

	  	{

	  	  sum += t.getElapsedTime();

	  	}

	  	catch (TimingException e)

	  	{

	  		System.err.println("ERROR, MUST STOP TIMER");

	  	}

	  }

	  

	  return sum / coalitions.size();

	}

	

	/**

	 * Add a uav representation to the collection of candidates prior to coalition formation

	 * 

	 * @param rep uav representation to add

	 */

	public void addUAVRepresentation(UAVRepresentation rep)

	{

		candidates.add(rep);

	}

	

	/**

	 * Adds a collection of uav representations to the collection of candidates prior to coalition formation

	 * 

	 * @param rep uav representation to add

	 */

	public void addUAVRepresentations(Collection col)

	{

		candidates.addAll(col);

	}



	/**

	 * Add a mission representation to the collection prior to coalition formation

	 * 

	 * @param rep uav representation to add

	 */

	public void addMissionRepresentation(MissionRepresentation rep)

	{

		missions.add(rep);

		missionFirstSolutionTimes.put(rep.getName(), Timing.getInstance());

	}	

	

	/**

	 * Add a collection of mission representations to the collection prior to coalition formation

	 * 

	 * @param col

	 */

	public void addMissionRepresentations(Collection col)

	{

		missions.addAll(col);

		

		Iterator iter = col.iterator();

		

		while (iter.hasNext())

		{

			MissionRepresentation rep = (MissionRepresentation) iter.next();

			missionFirstSolutionTimes.put(rep.getName(), Timing.getInstance());

		}

	}

	

	/**

	 * Form the coalitions. (separate method to allow for timing)

	 */

	public abstract void formCoalitions();

	

	/**

	 * Return the total number of combinations which were considered during run. 

	 *  

	 * @return total number of combinations which were considered during run

	 */

	public abstract BigInteger getNumCombosConsidered();

	

	/**

	 * Obtain the coalitions after forming.

	 * 

	 * @return List of Coalition objects.

	 */

	public List getCoalitions()

	{

		return coalitions;

	}

	

	/**

	 * Display the assigned coalitions and missions, as well as missions and candidates which were

	 * not assigned.

	 */

	public void display()

	{	

		// successfully formed coalitions

		if (coalitions.size() > 0)

			System.out.println("----------\nCOALITIONS\n----------\n");

			

		for (int i=0; i<coalitions.size(); i++)

		{

			Coalition coalition = (Coalition) coalitions.get(i);

			System.out.println("MISSION: " + coalition.getMissionRep().getName());

			for (int j=0; j<coalition.getUAVRepresentations().size(); j++)

			  System.out.println(" UAV: " + ((UAVRepresentation) coalition.getUAVRepresentations().get(j)).getName());

						 

		  Timing timing = (Timing) missionFirstSolutionTimes.get(coalition.getMissionRep().getName());

		  

		  try

		  {

			  System.out.println("First solution: " + Format.format(timing.getElapsedTime()));

		  }

		  catch(TimingException e)

		  {

		  	System.err.println("MUST STOP TIMER!");	

		  }

		  

			System.out.println("Number combinations considered: " + timing.getCombinationCount().toString());

			

			System.out.println();

		}

		

		// remaining missions

		if (missions.size() > 0)

			System.out.println("----------\nUNASSIGNED MISSIONS\n----------\n");	

				

		for (int i=0; i<missions.size(); i++)

		{

		  System.out.println("MISSION: " + ((MissionRepresentation) missions.get(i)).getName());

			Timing timing = (Timing) missionFirstSolutionTimes.get(((MissionRepresentation) missions.get(i)).getName());

			

			try

			{

			  System.out.println("Total time: " + Format.format(timing.getElapsedTime()));

			}

			catch (TimingException e)

			{

			  System.err.println("MUST STOP TIMER!");			

			}

			

			System.out.println("Number combinations considered: " + timing.getCombinationCount().toString());		  

		}

		 

		System.out.println();

		

		// remaining uavs

		if (candidates.size() > 0)

			System.out.println("----------\nUNASSIGNED UAVs\n----------\n");	

				

		for (int i=0; i<candidates.size(); i++)

			System.out.println("UAV: " + ((UAVRepresentation) candidates.get(i)).getName());

		 

		System.out.println();		

	}

		

	/**

	 * Display the sames of all the uavs in the potential coalition.

	 * 

	 * @param list collection of UAVRepresentation objects

	 */

	public void displayCoalitionNames(List list)

	{

		for (int i=0; i<list.size(); i++)

		{

			UAVRepresentation rep = (UAVRepresentation) list.get(i);

			System.out.println(" " + rep.getName());

		}

		

		System.out.println();

	}	



  /**

   * @return

   */

  public Map getMissionFirstSolutionTimes()

  {

    return missionFirstSolutionTimes;

  }

  

  /**

   * Return first solution time (sec) for the mission.  If more than one mission had coalitions formed,

   * pick one at random.

   *  

   * @return time to first solution for the mission, or -1.0 if error

   */

  public double getFirstSolutionTime()

  {

  	Set keys = missionFirstSolutionTimes.keySet();

		Iterator iter = keys.iterator();

		if (iter.hasNext() )

		{

			Object key = iter.next();

			Timing timing = (Timing) missionFirstSolutionTimes.get(key);

			try

			{

			  return timing.getElapsedTime();

			}

			catch (TimingException e)

			{

				e.printStackTrace();

			}

		}

		

		return -1.0;

  }

  



  /**

   * @return

   */

  public Timing getTotalTime()

  {

    return totalTime;

  }



  /**

   * @return

   */

  public int[] getQualityCost()

  {

    return qualityCost;

  }



  /**

   * @return

   */

  public double[] getQualityTime()

  {

    return qualityTime;

  }



}

