package yaes.application.uavaltclient.coalitionformation;



import java.util.List;



import yaes.application.uavaltclient.nav.NavUtil;



/**

 * Functions for computing coaltition cost as applicable to a mission, either as a whole or 

 * as component cost.

 * 

 * @author dejongp

 */

public class CostFunctions

{

	/**

	 * Determines whether or not a coalition has met the mission requirements.  Returns the cost 

	 * associated with this coalition according to the cost formula.  Cost is determined by

	 * summing the difference between the required capability for the mission and the actual

	 * capability of coalition.  Lower cost is better.

	 *  

	 * @param proposedCoalition list of uav reps containing capabilities

	 * @param missionRep mission containing the required capabilities

	 * @return cost associated with using this coalition (-1.0 if this is not a feasible coaltion)

	 */

	public static double cost(List proposedCoalition, MissionRepresentation missionRep)

	{

		double cost = 0.0;

		double componentCost;

				

		if (missionRep.getFirepower() > 0.0)

		{

			componentCost = firepowerCost(proposedCoalition, missionRep);

			

			//System.out.println("FIREPOWER COST " + componentCost);

			

			if (componentCost >= 0.0)

				cost += componentCost;

			else

				return -1.0;

		}

		

		if (missionRep.getNumBombs() > 0)

		{

			componentCost = bombCost(proposedCoalition, missionRep);



			if (componentCost >= 0)

				cost += (double) componentCost;

			else

				return -1.0;

		}

		 

		if (missionRep.getVisibility() > 0.0)

		{

			componentCost = visibilityCost(proposedCoalition, missionRep);

			if (componentCost >= 0.0)

				cost += componentCost;

			else

				return -1.0;

		}

		  

		if (missionRep.getMaxVelocity() > 0.0)

		{

			componentCost = velocityCost(proposedCoalition, missionRep);

			if (componentCost >= 0.0)

				cost += componentCost;

			else

				return -1.0;

		}

		  

		if (missionRep.getStealth() > 0.0)

		{

			componentCost = stealthCost(proposedCoalition, missionRep);

			

			//System.out.println("STEALTH COST " + componentCost);

			

			if (componentCost >= 0.0)

				cost += componentCost;

			else

				return -1.0;

		}

		

		// closer UAVs reduce cost

		cost += distanceCost(proposedCoalition, missionRep);

		

		// smaller coalitions reduce cost

		cost += coalitionSizeCost(proposedCoalition.size()); 

		

		return cost;

	}



	/**

	 * Determine distance component cost of coaltion.  The cost is determined based on the UAV which is furthest

	 * from the target.

	 * 

	 * @param proposedCoalition

	 * @param missionRep

	 * @return proposed coalition cost for distance component of mission.

	 */

  public static double distanceCost(List proposedCoalition, MissionRepresentation missionRep)

  {

  	double maxDist = 0.0;

  	

		for (int i=0; i<proposedCoalition.size(); i++)

		{

		  UAVRepresentation uav = (UAVRepresentation) proposedCoalition.get(i);

		  double distance = NavUtil.distance(uav.getCurrentPos(), missionRep.getTargetPos());

			if (distance > maxDist)

				maxDist = distance;

		}

		

		return maxDist / 10.0; //scale this down a little so distance isn't the most important factor

  }



  /**

   * Coalition size component of cost.

   * 

   * @param coalitionSize size of proposed coalition

   * @return coalition cost for size component

   */

  public static double coalitionSizeCost(int coalitionSize)

  {

  	return 2.0 * coalitionSize;

  }



	/**

	 * Determine firepower component cost of coaltion. A negative value indicates that the coalition

	 * is not feasible.

	 * 

	 * @param proposedCoalition

	 * @param missionRep

	 * @return proposed coalition cost for firepower component of mission.  A negative value indicates coalition is not feasible.

	 */

	public static double firepowerCost(List proposedCoalition, MissionRepresentation missionRep)

	{

		double firepower = 0.0;

				

		for (int i=0; i<proposedCoalition.size(); i++)

			firepower += ((UAVRepresentation) proposedCoalition.get(i)).getFirepower();



		return firepower - missionRep.getFirepower();

	}

	

	/**

	 * Determine bomb component cost of coalition.  A negative value indicates that the coalition is 

	 * not feasible.  Bomb cost is 10*number of bombs extra.

	 * 

	 * @param proposedCoalition

	 * @param missionRep

	 * @return proposed coalition cost for bomb component of mission.  A negative value indicates coalition is not feasible.

	 */

	public static int bombCost(List proposedCoalition, MissionRepresentation missionRep)

	{

		int numBombs = 0;

				

		for (int i=0; i<proposedCoalition.size(); i++)

		  numBombs += ((UAVRepresentation) proposedCoalition.get(i)).getNumBombs();



		return 10 * (numBombs - missionRep.getNumBombs());

	}

	

	/**

	 * Determine visibility component cost of coalition. A negative value indicates that the coalition

	 * is not feasible.

	 * 

	 * @param proposedCoalition

	 * @param missionRep

	 * @return proposed coalition cost for visibility component of mission.  A negative value indicates coalition is not feasible.

	 */

	public static double visibilityCost(List proposedCoalition, MissionRepresentation missionRep)

	{

		double maxVisibility = 0.0;

				

		for (int i=0; i<proposedCoalition.size(); i++)

		{

			double visibility = ((UAVRepresentation) proposedCoalition.get(i)).getVisibility();

			

			if (visibility > maxVisibility)

				maxVisibility = visibility;

		}

		 

		return maxVisibility - missionRep.getVisibility();

	}

	

	/**

	 * Determine velocity component cost of coalition. A negative value indicates that the coalition

	 * is not feasible.  The maxVelocity of the slowest UAV must equal or exceed the max required mission 

	 * velocity.

	 * 

	 * @param proposedCoalition

	 * @param missionRep

	 * @return proposed coalition cost for velocity component of mission.  A negative value indicates coalition is not feasible.

	 */

	public static double velocityCost(List proposedCoalition, MissionRepresentation missionRep)

	{

		double minVelocity = ((UAVRepresentation) proposedCoalition.get(0)).getMaxVelocity();

				

		for (int i=0; i<proposedCoalition.size(); i++)

		{

			double velocity = ((UAVRepresentation) proposedCoalition.get(i)).getMaxVelocity();

			

			if (velocity < minVelocity)

				minVelocity = velocity;				

		}



		return minVelocity - missionRep.getMaxVelocity();

	}

	

	/**

	 * Determine stealth component cost of coalition.  First determine if steath requirement is met using stealth formula.  

	 * If possible, cost is n.  Otherwise, cost is negative (coalition not feasible).  Higher stealth is better.

	 * 

	 * @param proposedCoalition

	 * @param missionRep

	 * @return proposed coalition cost for stealth component of mission.  A negative value indicates coalition is not feasible.

	 */

	public static double stealthCost(List proposedCoalition, MissionRepresentation missionRep)

	{

		double stealthSum = 0.0;

				

		for (int i=0; i<proposedCoalition.size(); i++)

			stealthSum += 10 - (((UAVRepresentation) proposedCoalition.get(i)).getStealth() / 10);

		

		double stealth = Math.max(1.0, 100 - stealthSum);

		//System.out.println("STEALTH=" + stealth + " sz=" + proposedCoalition.size());

	



//		if (stealth >= missionRep.getStealth())

//			return proposedCoalition.size(); 

//		

//		return -1;



    // this provides more useful information than simply "-1" (i.e., not feasible)		

		return stealth - missionRep.getStealth();

	}

		

		

		

//	/**

//	 * Determine stealth component cost of coalition.

//	 * 

//	 * @param proposedCoalition

//	 * @param missionRep

//	 * @return proposed coalition cost for stealth component of mission.  A negative value indicates coalition is not feasible.

//	 */

//	private double oldStealthCost(List proposedCoalition, MissionRepresentation missionRep)

//	{

//		double stealthProduct = ((UAVRepresentation) proposedCoalition.get(0)).getStealth();

//		double stealthSum = stealthProduct;

//				

//		for (int i=0; i<proposedCoalition.size(); i++)

//		{

//			double stealth = ((UAVRepresentation) proposedCoalition.get(i)).getStealth();

//			stealthProduct *= stealth;

//			stealthSum += stealth;

//		}

//		

//		//System.out.println("prod/sum: " + (stealthProduct / stealthSum));

//		

//		return  missionRep.getStealth() - stealthProduct / stealthSum;

//	}

		

}

