package yaes.application.uavaltclient.coalitionformation;



import java.math.BigInteger;

import java.util.ArrayList;

import java.util.List;



import yaes.application.uavaltclient.coalitionformation.timing.*;



/**

 * Determine coalition formation by using a depth-first iterative deepening algorithm.

 * 

 * PRUNE = FALSE

 * Determines coalition formation by considering all 2^n combinations, but in order

 * of coalition size.  If a solution is found for a certain size, the coaltion is

 * assigned to the mission and the algorithm re-starts for the next mission.  If

 * several solutions are found for the same size, the solution with the lowest

 * cost is selected.  The depth-first iterative deepening algorithm is used.

 * 

 * PRUNE = TRUE

 * As above, but apply heuristics to limit search size

 * 

 * @author dejongp

 */

public class DFIDCoalitionFormation extends CoalitionFormation

{

	private BigInteger numCombosConsidered = BigInteger.ZERO;

	private boolean pruneFlag;

	

	/**

	 * Constructor

	 * 

	 * @param pruneFlag true if pruning (heuristic application) is enabled, false otherwise

	 */

	public DFIDCoalitionFormation(boolean pruneFlag)

	{

		super();

		this.pruneFlag = pruneFlag;

	}

			

	/**

	 * For each mission: form the coalitions using a brute force method with

	 * iterative deepening (try all possibilities).

	 * Precondition: missions are sorted in the order they will be assigned.  

	 */

	public void formCoalitions()

	{

		totalTime.start();

		DepthFirstAlgorithm generator;

    

		List coalition = null;

		List unassignedMissions = new ArrayList();

		

		List leastExpensiveCoalition;

		int leastExpensiveCoalitionCost;

		

		int[] comboIndicies;



		while (missions.size() > 0 && candidates.size() > 0)

		{

			generator = new DFIDAlgorithm(candidates.size());   

			MissionRepresentation mRep = (MissionRepresentation) missions.get(0);

			Timing missionTiming = (Timing) missionFirstSolutionTimes.get(mRep.getName());

			missionTiming.start();

			

			if (qualityTimingFlag)

				startQualityTimers();

			

			leastExpensiveCoalition = null;

			leastExpensiveCoalitionCost = -1;

			int qualityIndex = 0;	

      

			comboIndicies = generator.nextCombo();

			

			while (comboIndicies != null)

			{

				int lastComboIndiciesLength = comboIndicies.length;

				

				coalition = new ArrayList();

	    	

				for (int i=0; i < comboIndicies.length; i++)

				{

					coalition.add(candidates.get(comboIndicies[i]));

				}

				 

				int cost = (int) Math.round(CostFunctions.cost(coalition, mRep));

        	

				if (cost >= 0 && ((cost < leastExpensiveCoalitionCost) || (leastExpensiveCoalitionCost == -1)))

				{

					

					if (qualityTimingFlag)

					{

						if (qualityIndex < MAX_QUALITY_TIMING_CNT)

						{

							qualityTimingObjects[qualityIndex].finish();

							try

							{

								qualityTime[qualityIndex] = qualityTimingObjects[qualityIndex].getElapsedTime();

							}

							catch (TimingException e)

							{

								e.printStackTrace();

							}

							qualityCost[qualityIndex] = cost;

						}

						else

							System.err.println("ERROR... NO MORE QUALITY TIMERS LEFT");					

						

						qualityIndex ++;

					}

					

					if (leastExpensiveCoalitionCost == -1)

					{

						//System.out.println("first solution mrep: " + mRep.getName() + " cost " + cost);

					  missionTiming.finish();

					}

					

					leastExpensiveCoalitionCost = cost;

					leastExpensiveCoalition = coalition;

					

					if (breakAfterFirstSolution)

						break;

				}

				

				if (pruneFlag && (cost < 0))

				{

					if ((mRep.getStealth() > 0.0 && CostFunctions.stealthCost(coalition, mRep) < 0) ||

							(mRep.getMaxVelocity() > 0.0 && CostFunctions.velocityCost(coalition, mRep) < 0))

					{

						generator.addBadBranch(comboIndicies);

					}

				}

				

				comboIndicies = generator.nextCombo();

				

				// TBD: changing this behavior.  Size is now a part of coalition cost.	

				

				// when combination length changes, pick a solution if one exists

//////				if (comboIndicies != null && (comboIndicies.length > lastComboIndiciesLength))

//////				{

//////					if (leastExpensiveCoalitionCost >= 0)

//////					{   

//////						coalitions.add(new Coalition(leastExpensiveCoalition, mRep));

//////						candidates.removeAll(leastExpensiveCoalition);

//////						missions.remove(mRep);

//////						break;

//////					}

//////				}

			}







			missions.remove(mRep);

			

			if (leastExpensiveCoalitionCost >= 0)

			{   

				coalitions.add(new Coalition(leastExpensiveCoalition, mRep));

				candidates.removeAll(leastExpensiveCoalition);

			}

			else

				unassignedMissions.add(mRep);

			    

			numCombosConsidered = numCombosConsidered.add(generator.getCombinationCount());

			missionTiming.addToCombinationCount(generator.getCombinationCount());

			

			

			

		}

    

		missions = unassignedMissions;

		totalTime.finish();

	}

	

	/**

	 * Return the total number of combinations which were considered during run. 

	 *  

	 * @return total number of combinations which were considered during run

	 */  

	public BigInteger getNumCombosConsidered()

	{

		return numCombosConsidered;

	}	

}

