package yaes.application.uavaltclient.coalitionformation;



import java.math.BigInteger;

import java.util.ArrayList;

import java.util.List;



import yaes.application.uavaltclient.coalitionformation.timing.*;



/**

 * Determines coalition formation using a Depth-First algorithm.

 * 

 * PRUNING = FALSE:

 * Determines coalition formation by considering all 2^n combinations.  The criteria for selection are

 * smallest coalition size, followed by least cost.  Uses the DepthFirst algorithm.

 * 

 * PRUNING = TRUE:

 * Determines coalition formation by considering all combinations using depth-first, 

 * but uses heuristics to short-circuit search paths which are guarenteed not to yield 

 * candidate coalitions.  Uses the depth-first algorithm.

 * 

 * @author dejongp

 */

public class DepthFirstCoalitionFormation extends CoalitionFormation

{

	private BigInteger numCombosConsidered = BigInteger.ZERO;

	private boolean pruneFlag;

	

	/**

	 * Constructor

	 * 

	 * @param pruneFlag true if pruning (heuristic application) is enabled, false otherwise

	 */

	public DepthFirstCoalitionFormation(boolean pruneFlag)

	{

		super();

		this.pruneFlag = pruneFlag;

	}

	

  /**

   * For each mission: form the coalitions using a brute force method (try all possibilities).

   * Precondition: missions are sorted in the order they will be assigned.  

   */

  public void formCoalitions()

  {

		totalTime.start();

		

    DepthFirstAlgorithm generator;

    

    List coalition = null;

		List unassignedMissions = new ArrayList();

		

		List leastExpensiveCoalition;

    int leastExpensiveCoalitionCost;

    int leastExpensiveCoalitionSize;

    

    int[] comboIndicies;

    int missionStartSize = missions.size();



    int bestCost = Integer.MAX_VALUE; 

		int qualityIndex = 0;

    

    while (missions.size() > 0 && candidates.size() > 0)

    {

      generator = new DepthFirstAlgorithm(candidates.size());   

			MissionRepresentation mRep = (MissionRepresentation) missions.get(0);		

			Timing missionTiming = (Timing) missionFirstSolutionTimes.get(mRep.getName());

			missionTiming.start();

			

			if (qualityTimingFlag)			

				startQualityTimers();



			leastExpensiveCoalition = null;

			leastExpensiveCoalitionCost = leastExpensiveCoalitionSize = -1;

      

			comboIndicies = generator.nextCombo();

			

			while (comboIndicies != null)

	    {

	    	coalition = new ArrayList();

	    	

				for (int i=0; i < comboIndicies.length; i++)

				{

					coalition.add(candidates.get(comboIndicies[i]));

				}

				 

				int cost = (int) Math.round(CostFunctions.cost(coalition, mRep));



        // non-negative cost means a solution has been found

				if (cost >= 0)

				{

					if (qualityTimingFlag && (cost < bestCost))

					{

						if (qualityIndex < MAX_QUALITY_TIMING_CNT)

						{

							qualityTimingObjects[qualityIndex].finish();

							try

							{

							  qualityTime[qualityIndex] = qualityTimingObjects[qualityIndex].getElapsedTime();

							}

							catch (TimingException e)

							{

								e.printStackTrace();

							}

							qualityCost[qualityIndex] = cost;

						}

						else

							System.err.println("ERROR... NO MORE QUALITY TIMERS LEFT");

						

						qualityIndex ++;

						//System.out.println("BETTER COST FOUND: old, new: " + bestCost + ", " + cost);

					  bestCost = cost;

					  

					}

					

					// record time of first solution

					if (leastExpensiveCoalitionSize == -1)

					{

						//System.out.println("mrep: " + mRep.getName());

					  missionTiming.finish();

					}

					

				  if ((leastExpensiveCoalitionSize == -1)  || // first possible coalition

				      (comboIndicies.length < leastExpensiveCoalitionSize) || // smaller coalition found

				      ((comboIndicies.length == leastExpensiveCoalitionSize) && (cost < leastExpensiveCoalitionCost))) // lower cost, equal size coaltion found

				  {

						leastExpensiveCoalitionCost = cost;

						leastExpensiveCoalitionSize = comboIndicies.length;

						leastExpensiveCoalition = coalition;

						if (breakAfterFirstSolution)

							break;

				  }

				}

				// negative cost... determine if heuristics apply

				else if (pruneFlag)

				{

					if ((mRep.getStealth() > 0.0 && CostFunctions.stealthCost(coalition, mRep) < 0) ||

							(mRep.getMaxVelocity() > 0.0 && CostFunctions.velocityCost(coalition, mRep) < 0))

					{

						generator.addBadBranch(comboIndicies);

					}

				}

								

				comboIndicies = generator.nextCombo();

	    }

    

			missions.remove(mRep);

			

      if (leastExpensiveCoalitionCost >= 0)

      {   

				coalitions.add(new Coalition(leastExpensiveCoalition, mRep));

				candidates.removeAll(leastExpensiveCoalition);

      }

      else

        unassignedMissions.add(mRep);

			    

			numCombosConsidered = numCombosConsidered.add(generator.getCombinationCount());

			missionTiming.addToCombinationCount(generator.getCombinationCount());

    }

    

    missions = unassignedMissions;

		totalTime.finish();

  }

  

	/**

	 * Return the total number of combinations which were considered during run. 

	 *  

	 * @return total number of combinations which were considered during run

	 */  

	public BigInteger getNumCombosConsidered()

	{

		return numCombosConsidered;

	}

}

