package yaes.application.uavaltclient.missionmgr;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import yaes.application.uavaltclient.nav.Pos;

/**
 * Defines the objectives of a squadron of UAVs.  Missions are kept by every UAV in the squadron.
 * 
 * @author dejongp
 */
public class Mission implements Cloneable
{
  public static final String DEFAULT_MISSION_FILE = "mission.xml";
	
	private List objectives = new ArrayList(); //Objective[]
	private int priority; // 1-5, 1==max priority (TBD: not sure what to use mission priority for yet)
	
	private int minUAVs; // min UAVs req'd to meet mission requirements
	// min number specialized UAVs req'd to meet mission requirements
	private int minFighters;
	private int minScouts;
	private int minBombers;
	
  //How many UAVs are being bootstrapped at startup.  Directory needs to know this.
	private int initialNumUAVs = -1;
	
  /**
   * Create an empty mission.
   */
  public Mission()
  {
  }
		
	/**
	 * Create a new mission using the xml configuration file.
	 */
	public Mission(String missionFile)
	{	
		// parse xml file
		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		
		try
		{
			DocumentBuilder builder = factory.newDocumentBuilder();
			Document doc = builder.parse(missionFile);
		  
			NodeList nl = doc.getElementsByTagName("mission");
			for (int i = 0; i < nl.getLength(); i++)
			{
				Element missionElement = (Element) nl.item(i);
				priority = Integer.parseInt(missionElement.getAttribute("priority"));
      
				NodeList nl2 = missionElement.getElementsByTagName("uavcount");

				Element countElement = (Element) nl2.item(0);
				minUAVs = Integer.parseInt(countElement.getAttribute("total"));
				minFighters = Integer.parseInt(countElement.getAttribute("fighter"));
				minBombers = Integer.parseInt(countElement.getAttribute("bomber"));
				minScouts = Integer.parseInt(countElement.getAttribute("scout"));	
				
				if (countElement.hasAttribute("initial"))
					initialNumUAVs = Integer.parseInt(countElement.getAttribute("initial"));
				
				NodeList nl3 = missionElement.getElementsByTagName("objective");
				for (int j = 0; j < nl3.getLength(); j++)
				{
					Objective obj = new Objective();
					Element objElement = (Element) nl3.item(j);
					
					obj.setName(objElement.getAttribute("name"));
					obj.setType(Objective.getObjectiveValue(objElement.getAttribute("type")));
					obj.setSubtype(Objective.getObjectiveValue(objElement.getAttribute("subtype")));
					
					if (objElement.hasAttribute("duration"))
						obj.setDuration(Integer.parseInt(objElement.getAttribute("duration")));					
				 
					obj.setPriority(Integer.parseInt(objElement.getAttribute("priority")));
				 
					NodeList nl4 = objElement.getElementsByTagName("position");

					Element posElement = (Element) nl4.item(0);
					obj.setPos1(new Pos(Double.parseDouble(posElement.getAttribute("x")),
															Double.parseDouble(posElement.getAttribute("y")),
															Double.parseDouble(posElement.getAttribute("z"))));
          
					posElement = (Element) nl4.item(1);
					if (posElement != null)
					{
						obj.setPos2(new Pos(Double.parseDouble(posElement.getAttribute("x")),
																Double.parseDouble(posElement.getAttribute("y")),
																Double.parseDouble(posElement.getAttribute("z"))));
					}
					
					NodeList nl5 = objElement.getElementsByTagName("capability");
					Element capabilityElement = (Element) nl5.item(0);
					
					if (! "".equals(capabilityElement.getAttribute("firepower")))
						obj.setFirepower(Integer.parseInt(capabilityElement.getAttribute("firepower")));	
					else if (! "".equals(capabilityElement.getAttribute("visibility")))
						obj.setVisibility(Integer.parseInt(capabilityElement.getAttribute("visibility")));
					else if (! "".equals(capabilityElement.getAttribute("velocity")))
						obj.setVelocity(Double.parseDouble(capabilityElement.getAttribute("velocity")));
					else if (! "".equals(capabilityElement.getAttribute("stealth")))
						obj.setStealth(Integer.parseInt(capabilityElement.getAttribute("stealth")));						
					
					addObjective(obj);
				}
			}
		}
		catch (Exception e)
		{
			System.err.println("Error parsing xml config file! " + e + ":" + e.getMessage() );		
		}
	}
	 
	/**
	 * Clone this Mission object.
	 */
	public Object clone()
	{
		try
		{
			Mission mission = (Mission) super.clone();
			
			if (objectives != null)
			{
				ArrayList newObjectives = new ArrayList();
				
				for (int i=0; i<objectives.size(); i++)
				{
					Objective obj = (Objective) objectives.get(i);
					newObjectives.add(obj.clone());
				}
				
				mission.objectives = newObjectives;
			}
				
			return mission;
		}
		catch (CloneNotSupportedException e)
		{
			throw new InternalError(e.toString());
		}  	
	}	 
	  

  /**
   * Merge a this mission with the supplied mission, ordering objectives by objective priority.
   * Place objectives which have already been completed at the beginning, below the 
   * currentObjectiveIndex pointer.
   * 
   * @param mission mission to merge with this mission.  'this' mission is currently executing, not param mission.
   * @param currentObjectiveIndex index in 'this' object indicating division between completed and uncompleted objectives.
   */
  public void mergeMission(Mission mission, int currentObjectiveIndex)
  {
		minUAVs += mission.getMinUAVs();
		minFighters += mission.getMinFighters();
		minScouts += mission.getMinScouts();
		minBombers += mission.minBombers;
		
		if (mission.priority < priority) priority = mission.priority;
		
		List completedObjectives = new ArrayList();
		
		for (int i=0; i<currentObjectiveIndex; i++)
		  completedObjectives.add(objectives.get(i));
		  
		objectives.removeAll(completedObjectives);

		objectives.addAll(mission.objectives);
		Collections.sort(objectives);
		
	  objectives.addAll(0, completedObjectives);
  }
  
  /**
   * Split the current mission in two, returning the second half as a new mission.
   * 
   * @return
   */
  public Mission splitMission()
  {		
		Mission newMission = new Mission();
		
		int newMissionObjCnt = objectives.size() / 2;
		for(int i=0; i<newMissionObjCnt; i++)
			newMission.addObjective(this.removeObjective(0));
		
		Collections.sort(objectives);
    Collections.sort(newMission.objectives);
    
    return newMission;	
  }
  
  /**
   * Call objective.toString on all objectives starting with the currently one 
   * (mgr.currentObjectiveIndex).
   * 
   * @param startingObj current objective index 
   * @return String displaying all objectives
   */
  public String getObjectiveList(int startingObj)
  {
  	String s = "";
  	
  	for (int i=startingObj; i<objectives.size(); i++)
  	{
  		Objective obj = (Objective) objectives.get(i);
  	  s += "OBJECTIVE #" + i + ":\r\n";
  	  s += obj.toString();	
  	}
  	
  	return s;
  }
  
  /**
   * Add an additional objective to the list of mission objectives.
   * TBD: This is public so that the GUI can add additional objectives.
   * 
   * @param o objective to be added
   */
  public void addObjective(Objective o)
  {
  	objectives.add(o);
  }
  
  /**
   * Obtain the mission objective located at the specified index in the list.
   * 
   * @param i index of desired objective within the list
   * @return Objective located at index i
   */
  public Objective getObjective(int i)
  {
  	try
  	{
  	  return (Objective) objectives.get(i);
  	}
  	catch (IndexOutOfBoundsException e)
  	{
  		Objective o = new Objective();
			// build default objective
			o.setName("DEFAULT OBJECTIVE");
			o.setType(Objective.getObjectiveValue("LOITER"));
			o.setDuration(999999999);//inf. loiter
			o.setPos1(MissionManager.DEFAULT_LOITER_POS);
			  		
  		return o;
  	}
  }
  
  /**
   * Remove (and return) an objective from the current objective list, shinking it by 1.
   * 
   * @param i index of objective to remove (0-based)
   * @return objective located at i
   */
  public Objective removeObjective(int i)
  {
  	return (Objective) objectives.remove(i);
  }
  
  /**
   * Obtain the current number of objectives.
   * 
   * @return number of objectives
   */
  public int getNumObjectives()
  {
  	return objectives.size();
  }


  /**
   * @return
   */
  public int getMinBombers()
  {
    return minBombers;
  }

  /**
   * @return
   */
  public int getMinFighters()
  {
    return minFighters;
  }

  /**
   * @return
   */
  public int getMinScouts()
  {
    return minScouts;
  }

  /**
   * @return
   */
  public int getMinUAVs()
  {
    return minUAVs;
  }
  
	/**
	 * @return
	 */
	public int getInitialNumUAVs()
	{
		return initialNumUAVs;
	}  

	/**
	 * For testing only.
	 * 
	 * @param args
	 */
	public static void main(String args[]) throws Exception
	{
		Mission mission1 = new Mission();
		Objective obj = new Objective();
		obj.setPos1(new Pos(1,2,3,4,5,6));
		obj.setName("start");
		mission1.objectives.add(obj);
		
		Mission mission2 = (Mission) mission1.clone();
		
/*
		// test XML parsing
		Mission mission = new Mission("mission1.xml");
		System.out.println("Num objectives orig mission: " + mission.getNumObjectives());
    
		Mission mission2 = new Mission("mission2.xml");
		mission2.priority = 2;
		mission2.minUAVs = 1;
		mission2.minFighters=0;
		mission2.minScouts=1;
		mission2.minBombers=0;
*/
		/*
		Objective o = new Objective();
		o.setDuration(1);
		o.setPriority(1);
		o.setType(Objective.RECON);
		mission2.addObjective(o);
    */
/*
		System.out.println("before merge:");
		System.out.println("mission1: " + mission.getObjectiveList(0));
		System.out.println("mission2: " + mission2.getObjectiveList(0));

    
    int currentObjInd = 2;
		mission.mergeMission(mission2, currentObjInd);

		System.out.println("after merge: mission1: " + mission.getObjectiveList(0));
    System.out.println("currentObjInd " + currentObjInd);
		System.out.println("only remaining objs: mission1: " + mission.getObjectiveList(currentObjInd));
*/
		/* test serialization
		String s = mission2.serialize();
		Mission mission2Dup = deserialize(s);
    
		System.out.println("Num objectives (1): " + mission2.getNumObjectives());
		System.out.println("Num objectives (2): " + mission2Dup.getNumObjectives());
		*/
	}


}
