package yaes.application.uavaltclient.missionmgr;

import java.util.Date;

import yaes.application.uavaltclient.agents.UAV;
import yaes.application.uavaltclient.message.MessageLibrary;
import yaes.application.uavaltclient.nav.NavUtil;
import yaes.application.uavaltclient.nav.Pos;
import yaes.application.uavaltclient.nav.SocPotField;
import yaes.application.uavaltclient.util.Props;

/**
 * Resolves mission objectives into UAV movements.  A Mission instance stays relatively static 
 * while the mission manager keeps track of the current state of the mission within a UAV.
 * 
 * @author dejongp
 */
public class MissionManager
{
	// mission modes
	public static final int MODE_NAV = 1; // navigate toward an objective of some sort
	public static final int MODE_FOLLOW = 2; // follow-the-leader behavior
	public static final int MODE_FORMATION = 3; // hold formation behind another UAV
  public static final int MODE_BOMBRUN = 4; // solo bomb run while rest of squadron "waits"
  public static final int MODE_WAIT = 5; // similar to loiter, mode for UAVs waiting on others

  // special modes
	public static final int MODE_SPF = 8;//(not currently a true mode)
	public static final int MODE_EMERGENCY = 9;

  // loitering modes
	public static final int MODE_LOITER_SINGLEPOINT = 21;
	public static final int MODE_LOITER_LAWNMOWER = 22;
	
  private int currentMode = MODE_FORMATION; //default to MODE_FORMATION on startup
  private int lastMode;
  private double emergencyTargetHdg;
  private double emergencyTargetAlt;
  
	private Mission mission;
	private Objective currentObjective;
	private long timeObjectiveReached = 0;
	private long durationAtObjective = 0; // seconds spent at current obj; compare to obj.duration
	private int currentObjectiveIndex = -1;
	private UAVMap map = new UAVMap();
  private SocPotField socPotField = new SocPotField();
  private int spfTime = 0; ///was static
	private String leader = null; //full aid of the leader (used for MODE_FOLLOW)  
  private int formationRank = -1;
  private String formationSide = "";


  // current aircraft status
  private Pos currentPosition = null;
  private Pos lawnmowerNextPos = null;
  
	// movement parameters resulting from computeNextMove(), which UAV should send to Simulation
	private double nextVel; //next velocity (absolute OpenGL coordinates/tick)
	private double nextHgt; //next height (relative map Y axis component)
	private double nextHdg; //next "orientation" (relative degrees. +:turn right, -:turn left)
	
	private UAV uav;
	
	// need somewhere to go when objectives are complete, emergency rondezvous, etc
	public static final Pos DEFAULT_LOITER_POS = new Pos(
	  Props.getDouble("DEFAULT_LOITER_POS_X"),
	  Props.getDouble("DEFAULT_LOITER_POS_Y"),
	  Props.getDouble("DEFAULT_LOITER_POS_Z"));

	private Objective defaultObjective = new Objective();
	private long initTime;
	private long now;
	private boolean isEmergency = false;
	private long cnt = 0;
	private Pos[] positions; //ac positions
	
	// holds bomb target
	private Pos waitPosition = null;
	private Pos bombPosition = null;
	private boolean bombsAwayFlag = false;

  // statistics recording for experiments
	private static boolean spfRecord = Props.getBoolean("SPF_RECORD");
	private static long totalSPFMillis = 0;
	private static double maxSPFForce = 0.0;
	private long spfStart = 0;
	
	// leader time
	private boolean ldrTimerSet = false;//true when ldrTime has been set
	private long ldrTimerSetAt;//when no leader detected and ldrTimeSetAt-now reaches LDR_TIME_LGTH ms, a new leader negotiation begins
  
  
	/**
	 * Constructor called when a new mission is received from user.
	 * 
	 * @param uav the UAV to which this MissionManager belongs
	 */
	public MissionManager(UAV uav, String missionFile)
	{
		mission = new Mission(missionFile);
		this.uav = uav;
		initTime = new Date().getTime();
		
		// build default objective
		defaultObjective.setName("DEFAULT OBJECTIVE");
		defaultObjective.setType(Objective.getObjectiveValue("LOITER"));
		defaultObjective.setDuration(999999999);//inf. loiter
		defaultObjective.setPos1(DEFAULT_LOITER_POS);  	
		
		nextObjective();
	}
	  
	/**
	 * Constructor called when a new mission is received from user.
	 * 
	 * @param uav the UAV to which this MissionManager belongs
	 */
	public MissionManager(UAV uav, Mission mission)
	{
		this.mission = mission;
		this.uav = uav;
		initTime = new Date().getTime();
		
		// build default objective
		defaultObjective.setType(Objective.getObjectiveValue("LOITER"));
		defaultObjective.setDuration(999999999);//inf. loiter
		defaultObjective.setPos1(DEFAULT_LOITER_POS);  	
		
		nextObjective();
	}	  
	  	  
	  
  /**
   * Based on current UAV position and all available mission information, put values for desired 
   * next move in nextVel, nextHgt, and nextHdg.
   */
  public void computeNextMove()
  {
  	cnt ++;
  	
  	if (cnt % 100 == 0)
  	{
			System.out.println("HEADING TOWARDS: " + currentObjective.getPos1().toString());
			System.out.println("uav " + uav.getName() + " mode: " + getCurrentMode());
  	}
		  	
		Date date = new Date();
		now = date.getTime();

		positions = map.getPositions(uav.getName()); //obtain positions of other aircraft
		checkLeadershipStatus();
		
		boolean isSPFThresholdExceeded = measureSPF();
    ///detectObstacles(); // currently disabled.
    
    if (currentMode == MODE_EMERGENCY)
    	emergencyHandling();
  	else if (currentMode == MODE_NAV)
  		navigationMode();
		else if (currentMode == MODE_BOMBRUN)
			bombrun();
		else if (currentMode == MODE_WAIT)
      waitMode();
  	else if (currentMode == MODE_LOITER_SINGLEPOINT)
      singlepointLoiter();			
		else if (currentMode == MODE_LOITER_LAWNMOWER)
      lawnmowerLoiter();	
  	else if (currentMode == MODE_FOLLOW)	 
      followTheLeader();
  	else if (currentMode == MODE_FORMATION)
      formationFlight();
  }
  
  
	/**
	 * Get the next objective from the Mission
	 */ 
	private void nextObjective()
	{
		System.out.println("\nTRANSITIONING TO NEXT OBJECTIVE\n");
		currentObjectiveIndex ++;
		
		// stay in MODE_NAV if this UAV is the leader;  otherwise, MODE_FOLLOW.
		if ((leader != null) && leader.equals(uav.getName()))
			currentMode = MODE_NAV;
	  
	  // TBD: NOT HITTING THIS CODE SINCE CURRENTLY FORMATION FLIGHT DOESN'T REGISTER COMPLETED OBJECTIVES...
	  ///else
	  	///currentMode = MODE_FOLLOW; //TBD: alter when formation flight it written

		durationAtObjective = 0;
		timeObjectiveReached = 0;	
		  	
		if (mission.getNumObjectives() > currentObjectiveIndex)
			currentObjective = mission.getObjective(currentObjectiveIndex);		
		else
		{
			if (spfRecord)
			{
				String results = "@@RESULTS: spfTime(s):" + (totalSPFMillis/1000) + " tot mis time(s): " + elapsedSeconds() + " maxSPF:" + maxSPFForce; 
				System.out.println(results);
				spfRecord = false;		  	
			}
					
			currentObjective = defaultObjective;
		}
		
		// notify squadron mates of new objective index
		// send out current position to other UAVs
		for (int i=0; i<uav.getSquadronMembers().length; i++)
			if (! uav.getName().equals(uav.getSquadronMembers()[i]))
				MessageLibrary.sendObjectiveNumMsg(uav.getName(), uav.getSquadronMembers()[i], new Integer(currentObjectiveIndex));		
		
	}
  
  /**
   * Instructs the mission manager to reload the current objective.  This is necessary after merging
   * missions, bomb runs, etc.
   */	
	public void reloadCurrentObjective()
	{
		currentObjective = mission.getObjective(currentObjectiveIndex);
		System.out.println("RELOADING CURRENT OBJECTIVE: " + currentObjective.getName() + " " + uav.getName());
	}
	
	/**
	 * Compute the next hdg command based on this UAV's limitations, put this value in nextHdg
	 * data member.
	 * 
	 * @param deltaHeading ideal turn if this UAV were physically capable
	 */
	private void assignNextHdg(double deltaHeading)
	{
		if (deltaHeading > uav.getMaxHdgDelta())
			nextHdg = uav.getMaxHdgDelta();
		else if (deltaHeading < -uav.getMaxHdgDelta())
			nextHdg = -uav.getMaxHdgDelta();
		else
			nextHdg = deltaHeading;
		
	}
	
	/**
	 * Compute the next height command based on this UAV's limitations, put this value in nextHgt
	 * data member.
	 * 
	 * @param deltaHeight ideal height change if this UAV were physically capable
	 */	
	private void assignNextHeight(double deltaHeight)
	{
		if (deltaHeight > NavUtil.NAV_HEIGHT_TOLERANCE)
		{
			if (deltaHeight > uav.getMaxAltDelta())
				nextHgt = uav.getMaxAltDelta();
			else
				nextHgt = deltaHeight;
		}
		else if (deltaHeight < -NavUtil.NAV_HEIGHT_TOLERANCE)
		{
			if (deltaHeight < -uav.getMaxAltDelta())
				nextHgt = -uav.getMaxAltDelta();
			else
				nextHgt = deltaHeight;
		}
		else
		  nextHgt = 0.0;
	}
	
	  
  /**
   * @param pos
   */
  public void setCurrentPosition(Pos pos)
  {
    currentPosition = pos;
  }

  /**
   * @return
   */
  public double getNextHdg()
  {
    return nextHdg;
  }

  /**
   * @return
   */
  public double getNextHgt()
  {
    return nextHgt;
  }

  /**
   * @return
   */
  public double getNextVel()
  {
    return nextVel;
  }

	/**
	 * @return
	 */
	public UAVMap getMap()
	{
		return map;
	}

  /**
   * @param string
   */
  public void setLeader(String string)
  {
    leader = string;
		ldrTimerSet = false;//true when ldrTime has been set
		ldrTimerSetAt = 0l; 
  }

  /**
   * @return
   */
  public int getCurrentMode()
  {
    return currentMode;
  }

  /**
   * @return
   */
  public Mission getMission()
  {
    return mission;
  }

  /**
   * @return
   */
  public Objective getCurrentObjective()
  {
    return currentObjective;
  }

  /**
   * @param i
   */
  public void setCurrentMode(int i)
  {
  	lastMode = currentMode;
    currentMode = i;
  }

  /**
   * @return a clone of current UAV position
   */
  public Pos getCurrentPosition()
  {
    return (Pos) currentPosition.clone();
  }
  
  /**
   * 
   * @return elapsed time in seconds between MissionManager initialization and now.
   */
  public long elapsedSeconds()
  {
  	long curTime = new Date().getTime();
  	return (curTime - initTime) / 1000;
  }

  /**
   * @param i
   */
  public void setFormationRank(int i)
  {
    formationRank = i;
  }

  /**
   * @param string
   */
  public void setFormationSide(String string)
  {
    formationSide = string;
  }

  /**
   * @return
   */
  public int getCurrentObjectiveIndex()
  {
    return currentObjectiveIndex;
  }

  /**
   * @return
   */
  public long getCnt()
  {
    return cnt;
  }

  /**
   * @param pos
   */
  public void setBombPosition(Pos pos)
  {
    bombPosition = pos;
  }

  /**
   * @return
   */
  public boolean isBombsAwayFlag()
  {
    return bombsAwayFlag;
  }

  /**
   * @param b
   */
  public void setBombsAwayFlag(boolean b)
  {
    bombsAwayFlag = b;
  }
  
  /**
   * Load previous state and continue with mission after an event such as a bomb run.
   */
  public void resumeMission()
  {
		setBombsAwayFlag(false);
		setBombPosition(null);
		setWaitPosition(null);  	
  	currentMode = lastMode;
		reloadCurrentObjective();
  }

  /**
   * @param pos
   */
  public void setWaitPosition(Pos pos)
  {
    waitPosition = pos;
  }

	/**
	 * Obtain the current time in millis.
	 * 
	 * @return current time in millis
	 */
	private static long getTimestamp()
	{
		Date date = new Date();
		return date.getTime();
	}
	
  /**
   * @param i
   */
  public void setCurrentObjectiveIndex(int i)
  {
    currentObjectiveIndex = i;
    currentObjective = mission.getObjective(i);
    //System.out.println("UAV " + uav.getName() + " sets coi to " + i + "\nobjective " + currentObjective.toString());
  }

  /**
   * Perform a leader negotiation if necessary.
   */
	private void checkLeadershipStatus()
	{
		// detect missing leader, initiate ldr neg if nec
		// (allow some some time for missions, modes, and leaders to be set before initiating ldr neg)
		if (leader == null && (currentMode == MODE_FOLLOW || currentMode == MODE_FORMATION))
		{
			if (! ldrTimerSet)
			{
				//System.out.println("SETTING TIMER FOR UAV " + uav.getName());
				ldrTimerSet = true;
				ldrTimerSetAt = now;
			}
			else if (now - ldrTimerSetAt >= Props.getInt("LDR_TIMER_LGTH"))
			{
				//System.out.println(" --> " + uav.getName() + ": no leader specified; Initiating leader negotiation.");
				uav.initiateLeaderNegotiation();
								
				ldrTimerSet = false;
				ldrTimerSetAt = 0l;
			}
		}		
	}

  /**
   * Perform emergency handling to avoid obstacle.
   */
  private void emergencyHandling()
  {
		if (currentPosition.getY() >= emergencyTargetAlt)
		{
			isEmergency = false;
			currentMode = lastMode;
			System.out.println("TRANSITIONING OUT OF EMERGENCY HANDLING");
		}
		else
		{
			assignNextHeight(emergencyTargetAlt);
        
			double deltaHeading = NavUtil.deltaHeading(currentPosition.getHdg(), emergencyTargetHdg);
			if (Math.abs(deltaHeading) > NavUtil.NAV_HEADING_TOLERANCE)
				assignNextHdg(deltaHeading);
			else 
				assignNextHdg(0.0);
					
			// slow to avoid hitting obstacle, then accelerate once turned away
			if (Math.abs(deltaHeading) >= 130.0)
				nextVel = 0.1;//slow as much as possible
			else
				nextVel = uav.getMaxVelocity();
			  					
			System.out.println(" EMER: assigning hgt:" + emergencyTargetAlt + " hdg:" + deltaHeading + " uav:" + currentPosition.toString());					
		}  	
  }
  
  /**
   * Navigate toward a target.
   *
   */
  private void navigationMode()
  {
		// end of nav mode;  determine what mode comes next
		if (NavUtil.isPosReached(currentPosition, currentObjective.getPos1()))
		{
			System.out.println("NAV POINT REACHED");
			timeObjectiveReached = now;
			
			if (currentObjective.getType() == Objective.LOITER)
			{
				if (currentObjective.getSubtype() == Objective.SINGLEPOINT)
				{
					currentMode = MODE_LOITER_SINGLEPOINT;
				}
				else if (currentObjective.getSubtype() == Objective.LAWNMOWER)
				{
					currentMode = MODE_LOITER_LAWNMOWER;
				}
			}
		}
		// still navigating towards target
		else
		{		
			assignNextHeight(NavUtil.deltaHeight(currentPosition, currentObjective.getPos1()));
				
				
			nextVel = uav.getCruiseVelocity(); //TBD: when leading, set this to the vel of the slowest follower
        
			double deltaHeading = NavUtil.deltaHeading(currentPosition, currentObjective.getPos1());
			if (Math.abs(deltaHeading) > NavUtil.NAV_HEADING_TOLERANCE)
				assignNextHdg(deltaHeading);
			else
				assignNextHdg(0.0);

		}  	
  }
  
  /**
   * Learn to stop worrying and love the bomb.
   *
   */
  private void bombrun()
  {
		if (NavUtil.isPosReached(currentPosition, bombPosition) && !bombsAwayFlag)
		{
			System.out.println("BOMBS AWAY!!!");
			setBombsAwayFlag(true);
		}
		// still navigating towards target
		else
		{		
			assignNextHeight(NavUtil.deltaHeight(currentPosition, bombPosition));
			nextVel = uav.getCruiseVelocity();
			double deltaHeading = NavUtil.deltaHeading(currentPosition, bombPosition);
			if (Math.abs(deltaHeading) > NavUtil.NAV_HEADING_TOLERANCE)
				assignNextHdg(deltaHeading);
			else 
				assignNextHdg(0.0);					
		}  	
  }
  
  /**
   * Wait for a bomb run to be carried out by another UAV.
   *
   */
  private void waitMode()
  {
		if (cnt % 10 == 0)
			System.out.println(uav.getName() + " MODE: WAIT");
  		
		assignNextHeight(NavUtil.deltaHeight(currentPosition, waitPosition));
		nextVel = uav.getCruiseVelocity();
        
		double deltaHeading = NavUtil.deltaHeading(currentPosition, waitPosition);
			
		if (Math.abs(deltaHeading) > NavUtil.NAV_HEADING_TOLERANCE)
			assignNextHdg(deltaHeading);
		else 
			assignNextHdg(0.0);  	
  }
  
  /**
   * Loiter around a single point in space.
   *
   */
  private void singlepointLoiter()
  {
		durationAtObjective = (now - timeObjectiveReached) / 1000;
			
		if (cnt % 2 == 0)
			System.out.println("SINGLEPOINT LOITER, t=" + durationAtObjective);
  			
		if (durationAtObjective > currentObjective.getDuration())
		{
			nextObjective();
			return;
		}			
			
		assignNextHeight(NavUtil.deltaHeight(currentPosition, currentObjective.getPos1()));
		nextVel = uav.getCruiseVelocity(); //TBD: when leading, set this to the vel of the slowest follower
        
		double deltaHeading = NavUtil.deltaHeading(currentPosition, currentObjective.getPos1());
		if (Math.abs(deltaHeading) > NavUtil.NAV_HEADING_TOLERANCE)
			assignNextHdg(deltaHeading);
		else 
			assignNextHdg(0.0);	  	
  }
  
  /**
   * Loiter within a rectangular area.
   *
   */
  private void lawnmowerLoiter()
  {
		durationAtObjective = (now - timeObjectiveReached) / 1000;
			
		if (cnt % 2 == 0)
			System.out.println("LAWNMOWER LOITER, t=" + durationAtObjective);
  			
		if (durationAtObjective > currentObjective.getDuration())
		{
			nextObjective();
		}
		else
		{	
			// first time through loop
			if (lawnmowerNextPos == null)
			{
				System.out.println("LAWNMOWER: Setting first point");
				lawnmowerNextPos = (Pos) currentObjective.getPos1().clone();
				lawnmowerNextPos.setZ(currentObjective.getPos2().getZ());
			}
				
			if (NavUtil.isPosReached(currentPosition, lawnmowerNextPos))
			{
				if (NavUtil.isPosReached(currentPosition, currentObjective.getPos2()))
				{
					System.out.println("LAWNMOWER: Reached last point; starting over");
					lawnmowerNextPos = (Pos) currentObjective.getPos1().clone();
				}
					
				// we are closer to pos1 Z axis... switch to pos2 Z axis for next leg
				else if ( Math.abs(currentPosition.getZ() - currentObjective.getPos1().getZ()) <
									Math.abs(currentPosition.getZ() - currentObjective.getPos2().getZ()) )
				{
					System.out.println("LAWNMOWER: Reached pos1.Z, heading for pos2.Z");
					lawnmowerNextPos.setZ(currentObjective.getPos2().getZ());
				}
				// we are closer to pos2 y axis... switch to pos1 Z axis, and x axis += LAWNMOWER_X_LEG
				else
				{
					System.out.println("LAWNMOWER: Reached pos2.Z, heading for X+=deltaX,pos1.Z");
					lawnmowerNextPos.setZ(currentObjective.getPos1().getZ());
					lawnmowerNextPos.setX(lawnmowerNextPos.getX() + Props.getDouble("LAWNMOWER_X_LEG"));	
					if (lawnmowerNextPos.getX() > currentObjective.getPos2().getX())
					{
						System.out.println("  LAWNMOWER: Max X reached; using pos2.X");
						lawnmowerNextPos.setX(currentObjective.getPos2().getX());
					}
				}
			}
				
			assignNextHeight(NavUtil.deltaHeight(currentPosition, lawnmowerNextPos));
			nextVel = uav.getCruiseVelocity(); //TBD: when leading, set this to the vel of the slowest follower
        
			double deltaHeading = NavUtil.deltaHeading(currentPosition, lawnmowerNextPos);
			if (Math.abs(deltaHeading) > NavUtil.NAV_HEADING_TOLERANCE)
				assignNextHdg(deltaHeading);
			else 
				assignNextHdg(0.0);
				
		}  	
  }
  
  /**
   * Follow another UAV.
   *
   */
  private void followTheLeader()
  {
		if (leader != null)
		{
			Pos leaderPos = map.getPosition(leader);
        
			try
			{
				assignNextHeight(NavUtil.deltaHeight(currentPosition, leaderPos));
          
				// TBD: could add "catch-up" velocity here if distance is great          
				if (leaderPos.getV() <= uav.getMaxVelocity())
					nextVel = leaderPos.getV();
				else
					nextVel = uav.getMaxVelocity();
          
				double deltaHeading = NavUtil.deltaHeading(currentPosition, leaderPos);
				if (Math.abs(deltaHeading) > NavUtil.NAV_HEADING_TOLERANCE)
					assignNextHdg(deltaHeading);
				else 
					assignNextHdg(0.0);				  	
			}
			catch(NullPointerException e)
			{
				System.out.println("CANNOT GET currentPosition / leaderPos");          
			}
		}  	
  }
  
  /**
   * Maintain a fixed distance from another uav.
   *
   */
  private void formationFlight()
  {
		//if (cnt % 20 == 0)
			//System.out.println(uav.getTheBondAgent().getLocalName() + " MODE: FORMATION: " + 
		//  	formationSide + " " + formationRank + " (following " + leader + ")");

		if ((leader != null) && ((cnt % Props.getInt("FORMATION_HDG_RECOMPUTE")) == 0)) //will have to tune % value
		{
			Pos leaderPos = map.getPosition(leader);
			if (leaderPos == null) return; // addresses bootstrap issue on startup
				
			// pos 0-3: leftX, leftY, rightX, rightY respectively (Y==Z in sim coords)
			double[] coords = NavUtil.getFormationLocation(leaderPos.getX(), 
																										 leaderPos.getZ(), 
																										 leaderPos.getHdg(), 
																										 Props.getInt("FORMATION_DISTANCE_BEHIND") * formationRank);
				
			double x,y;
			if (formationSide.equals("left"))
			{
				x = coords[0];
				y = coords[1];
			}
			else
			{
				x = coords[2];
				y = coords[3];
			}
				
			try
			{
				assignNextHeight(NavUtil.deltaHeight(currentPosition, leaderPos));
					
				// catch-up velocity
				if (NavUtil.distance(currentPosition, leaderPos) > 
						(NavUtil.distance(x, y, leaderPos.getX(), leaderPos.getZ()) + NavUtil.NAV_FORMATION_DISTANCE_TOLERANCE))
				{
//					System.out.println(uav.getTheBondAgent().getLocalName() + " attempting to catch up!");
					nextVel = uav.getMaxVelocity();
				}
				else if (leaderPos.getV() <= uav.getMaxVelocity())
					nextVel = leaderPos.getV();
				else
					nextVel = uav.getMaxVelocity();
          
				double deltaHeading = NavUtil.deltaHeading(currentPosition, new Pos(x, currentPosition.getHdg(), y));
					
				if (Math.abs(deltaHeading) > NavUtil.NAV_HEADING_TOLERANCE)
					assignNextHdg(.5 * deltaHeading); //TBD: tmp!  This dampens the inexplicable oscillation
				else 
					assignNextHdg(0.0);						
					
				/*	
				System.out.println(uav.getTheBondAgent().getLocalName() + ".  " +
					"Leader:" + leaderPos.getX() + "," + leaderPos.getZ() +
					", Form tgt:" + x + "," + y); 
				*/
					
			}
			catch(NullPointerException e)
			{
				System.out.println("CANNOT GET currentPosition / leaderPos");          
			}
		}  	
  }
  
  /**
   * Measure SPF, determine if threshold has been exceeded.  If exceeded, determine AC reaction to
   * avoid collision with other vehicles.
   * 
   * @return true if threshold was exceeded and SPF acted upon, false otherwise
   */
  private boolean measureSPF()
	{
		// don't bother checking SPF for solitary uavs
		if (positions == null || positions.length == 0)
			return false;
			
		if (spfTime == 0)
		{
			socPotField.computeField(currentPosition, positions);
       
			if ((spfRecord && elapsedSeconds() > Props.getInt("SPF_ENABLE_AFTER_SECONDS") && socPotField.getForce() > maxSPFForce))
				maxSPFForce = socPotField.getForce();
					       
			// TBD: elapsed time check is a hack to avoid aircraft freaking out since they all come into being 
			// occupying the same space.  This turns off SPF for a certain amount of time.
			if ((socPotField.getForce() > SocPotField.SIGNFICANT_FORCE) && 
					(elapsedSeconds() > Props.getInt("SPF_ENABLE_AFTER_SECONDS")))
			{
				spfTime ++;
				  
				if (spfRecord)
					spfStart = getTimestamp();
					      	
				System.out.println(now + " " + uav.getName() + " me: " + 
					currentPosition.toString() + " them:" + positions[0].toString() + "\n " +
					"\n -->SPF: " + socPotField.getString()); 
					      	
				System.out.println(uav.getName() + " MODE: SPF INIT");
				nextVel = uav.getMaxVelocity();
				nextHgt = 0.0; //TBD: tmp.  Use socField.Y()?  Or value from nav?  Or is no change ok?  
				assignNextHdg(socPotField.getHdg() - currentPosition.getHdg() );
			}
			else
				return false;
		}
		// must be sitting in SPF for MAX_SPF_TIME ticks
		else
		{
			spfTime ++;
				
			System.out.println(uav.getName() + " MODE: SPF SIT " + spfTime);
			nextVel = uav.getMaxVelocity();
			nextHgt = 0.0; //TBD: tmp.  Use socField.Y()?  Or value from nav?  Or is no change ok?  
			assignNextHdg(socPotField.getHdg() - currentPosition.getHdg() );
			if (spfTime == Props.getInt("MAX_SPF_TIME"))
				spfTime = 0;
					
			if (spfRecord)
					totalSPFMillis += getTimestamp() - spfStart;
		}
		
		return true;
	}
	
	
	/**
	 * Determine if this uav is in imminent danger of colliding with a stationary obstacle.
	 *
	 */
	private void detectObstacles()
	{ 
		if ((uav.elapsedTime() > 15) && (! isEmergency))
		{
			int acX = (int) Math.round(currentPosition.getX());
			int acZ = (int) Math.round(currentPosition.getZ());
			int startX, limitX, startZ, limitZ, xAdj, xAdjSum = 0;
			double legDouble = Math.sqrt(((double) (Props.getInt("OBSTACLE_AVOIDANCE_DISTANCE_AHEAD") * Props.getInt("OBSTACLE_AVOIDANCE_DISTANCE_AHEAD"))) / 2);
			int leg = (int) Math.round(legDouble);
      
			// heading N
			if ((currentPosition.getHdg() > 337.5) || (currentPosition.getHdg() <= 22.5))
			{
				startX = acX - (Props.getInt("OBSTACLE_AVOIDANCE_LATERAL_RANGE") / 2);
				limitX = acX + (Props.getInt("OBSTACLE_AVOIDANCE_LATERAL_RANGE") / 2);
				startZ = acZ - Props.getInt("OBSTACLE_AVOIDANCE_DISTANCE_AHEAD");
				limitZ = acZ;
				xAdj = 0;
			}
			// heading NE
			else if ((currentPosition.getHdg() > 22.5) && (currentPosition.getHdg() <= 67.5))
			{
				startX = (acX + leg) - (Props.getInt("OBSTACLE_AVOIDANCE_LATERAL_RANGE") / 2);
				limitX = (acX + leg) + (Props.getInt("OBSTACLE_AVOIDANCE_LATERAL_RANGE") / 2);
				startZ = acZ - leg;
				limitZ = acZ;
				xAdj = -1;
			}
			// heading E
			else if ((currentPosition.getHdg() > 67.5) && (currentPosition.getHdg() <= 112.5))
			{
				startX = acX;
				limitX = acX + Props.getInt("OBSTACLE_AVOIDANCE_DISTANCE_AHEAD");
				startZ = acZ - (Props.getInt("OBSTACLE_AVOIDANCE_LATERAL_RANGE") / 2);
				limitZ = acZ + (Props.getInt("OBSTACLE_AVOIDANCE_LATERAL_RANGE") / 2);
				xAdj = 0;
			}
			// heading SE
			else if ((currentPosition.getHdg() > 112.5) && (currentPosition.getHdg() <= 157.5))
			{
				startX = acX - (Props.getInt("OBSTACLE_AVOIDANCE_LATERAL_RANGE") / 2);
				limitX = acX + (Props.getInt("OBSTACLE_AVOIDANCE_LATERAL_RANGE") / 2);
				startZ = acZ;
				limitZ = acZ + leg;
				xAdj = 1;
			}
			// heading S
			else if ((currentPosition.getHdg() > 157.5) || (currentPosition.getHdg() <= 202.5))
			{
				startX = acX - (Props.getInt("OBSTACLE_AVOIDANCE_LATERAL_RANGE") / 2);
				limitX = acX + (Props.getInt("OBSTACLE_AVOIDANCE_LATERAL_RANGE") / 2);
				startZ = acZ;
				limitZ = acZ + Props.getInt("OBSTACLE_AVOIDANCE_DISTANCE_AHEAD");
				xAdj = 0;
			}
			// heading SW
			else if ((currentPosition.getHdg() > 202.5) && (currentPosition.getHdg() <= 247.5))
			{
				startX = acX - (Props.getInt("OBSTACLE_AVOIDANCE_LATERAL_RANGE") / 2);
				limitX = acX + (Props.getInt("OBSTACLE_AVOIDANCE_LATERAL_RANGE") / 2);
				startZ = acZ;
				limitZ = acZ + leg;
				xAdj = -1;
			}
			// heading W
			else if ((currentPosition.getHdg() > 247.5) && (currentPosition.getHdg() <= 292.5))
			{
				startX = acX - Props.getInt("OBSTACLE_AVOIDANCE_DISTANCE_AHEAD");
				limitX = acX;
				startZ = acZ - (Props.getInt("OBSTACLE_AVOIDANCE_LATERAL_RANGE") / 2);
				limitZ = acZ + (Props.getInt("OBSTACLE_AVOIDANCE_LATERAL_RANGE") / 2);
				xAdj = 0;
			}
			// heading NW
			else // if ((currentPosition.getHdg() > 292.5) && (currentPosition.getHdg() <= 337.5))
				{
				startX = (acX - leg) - (Props.getInt("OBSTACLE_AVOIDANCE_LATERAL_RANGE") / 2);
				limitX = (acX - leg) + (Props.getInt("OBSTACLE_AVOIDANCE_LATERAL_RANGE") / 2);
				startZ = acZ - leg;
				limitZ = acZ;
				xAdj = 1;
			}

			for (int z = startZ; z <= limitZ; z++)
			{
				if (z != startZ)
					xAdjSum += xAdj;
				for (int x = startX + xAdjSum; x <= limitX; x++)
				{
					double obstacleHeight = map.getHeight(x, z);
					//System.out.println("   scan " + x + "," + z + " : " + obstacleHeight);
					if (obstacleHeight >= currentPosition.getY())
					{
						lastMode = currentMode;
						currentMode = MODE_EMERGENCY;
						emergencyTargetHdg = currentPosition.getHdg() + 180.0;
						if (emergencyTargetHdg >= 360.0)
							emergencyTargetHdg -= 360.0;
						emergencyTargetAlt = currentPosition.getY() + Props.getInt("EMERGENCY_ASCEND_DISTANCE");
						System.out.print("WARNING: OBSTACLE NEAR");
						System.out.print(" current x,z: " + currentPosition.getX() + "," + currentPosition.getZ() + " Y:" + currentPosition.getY());
						System.out.println(" obstacle x,z: " + x + "," + z + " alt:" + obstacleHeight);
						isEmergency = true;
						break;
					}
				}
				if (isEmergency)
					break;
			}
		}		
	}
}
  