package yaes.application.uavaltclient.naming;

import java.util.ArrayList;

/**
 * Models extended name (unique agent name).
 * 
 * @author dejongp
 */
public class ExtendedName
{
	private ArrayList identitySet;
	private int splitNumber;
	private ArrayList candidacySet; // a list of lists
	
	private static String nextIdentity = "";
	
	/**
	 * Construct a new extened name in its initial state (no split, no candidacy, etc), assigning the
	 * next identity in the sequence.
	 */
	public ExtendedName()
	{
		identitySet = new ArrayList();
		identitySet.add(getNextIdentity());
		splitNumber = 0;
		candidacySet = new ArrayList();
	}
	
  /**
   * Construct a new extended name based on the supplied parameters.
   * 
   * @param identitySet
   * @param splitNumber
   * @param candidacySet
   */
  private ExtendedName(ArrayList identitySet, int splitNumber, ArrayList candidacySet)
  {
    this.identitySet = identitySet;
    this.splitNumber = splitNumber;
    this.candidacySet = candidacySet;
  }
  
	
	/**
	 * Deep clone this object, and return its new clone.
	 * 
	 * @return Object newly cloned object
	 */
	public Object clone()
	{
		// shallow copy ok here
    ArrayList newIdentitySet = (ArrayList) identitySet.clone();
    
    ArrayList newCandidacySet = new ArrayList();
		for (int i=0; i<candidacySet.size(); i++)
		{
			ArrayList newCandidacy = (ArrayList) ((ArrayList) candidacySet.get(i)).clone();
			newCandidacySet.add(newCandidacy);
		}
    
    return new ExtendedName(newIdentitySet, splitNumber + 1, newCandidacySet);
	}
		
	/**
	 * Return the visual version of this object; e.g, ({A}, 0, {}) 
	 * 
	 * @return String version of this object
	 */
	public String toString()
	{
		String s = "({";
		for(int i=0; i<identitySet.size(); i++)
		{
			s += (String) identitySet.get(i);
			if (i < (identitySet.size() - 1)) s += ","; 
		}
		s += "}," + splitNumber + ",{";
		
		for (int i=0; i<candidacySet.size(); i++)
		{
			ArrayList candidacy = (ArrayList) candidacySet.get(i);
			s += "(";
			for (int j=0; j<candidacy.size(); j++)
			{
				if ((candidacy.get(j)).getClass().getName().equals("java.lang.String"))
					s += (String) candidacy.get(j);
				else if ((candidacy.get(j)).getClass().getName().equals("java.lang.Integer"))
					s += ((Integer) candidacy.get(j)).toString();					

			  if (j < (candidacy.size() - 1)) s += ",";	
			}
			s += ")";
			if (i < (candidacySet.size() - 1)) s += ","; 
		}
		s += "})";
		
		return s;
	}
	
	
	/**
	 * The entity this name was assigned to has been split.  Generate two resulting names based on
	 * this name according to the "Split rule."
	 * 
	 * @return List containing two new ExtendedNames resulting from application of "Split rule"
	 */
	public ArrayList split()
	{
		// first extended name
		ExtendedName name1 = (ExtendedName) this.clone();
		name1.splitNumber = this.splitNumber + 1;
		
		// second extended name
		ArrayList idset = new ArrayList();
		idset.add(getNextIdentity());
		ArrayList cset = new ArrayList();
		
		for (int i=0; i<identitySet.size(); i++)
		{
			ArrayList newCandidacy = new ArrayList();
			newCandidacy.add(identitySet.get(i));
			newCandidacy.add(new Integer(splitNumber));
		  cset.add(newCandidacy);
		}

		for (int i=0; i<candidacySet.size(); i++)
		{
			ArrayList newCandidacy = (ArrayList) ((ArrayList) candidacySet.get(i)).clone();
			newCandidacy.add(new Integer(0));
			cset.add(newCandidacy);
		}
		
		ExtendedName name2 = new ExtendedName(idset, 0, cset);
		ArrayList ret = new ArrayList();
		ret.add(name1);
		ret.add(name2);
		return ret;
	}
	
	/**
	 * Generate a new ExtendedName by applying the merge rule to provided ExtendedNames name1 and 
	 * name2.
	 * 
	 * @param name1 ExtenedName to merge with name2
	 * @param name2 ExtenedName to merge with name1
	 * @return ExtendedName resulting from merger of name1 + name2
	 */
	public static ExtendedName merge(ExtendedName name1, ExtendedName name2)
	{
		ArrayList idset = (ArrayList) name1.identitySet.clone();
		idset.addAll(name2.identitySet);
		
		ArrayList cset = new ArrayList();
		
		for (int i=0; i<name1.candidacySet.size(); i++)
		{
			ArrayList newCandidacy = (ArrayList) ((ArrayList) name1.candidacySet.get(i)).clone();
			cset.add(newCandidacy);
		}
		
		for (int i=0; i<name2.candidacySet.size(); i++)
		{
			ArrayList newCandidacy = (ArrayList) ((ArrayList) name2.candidacySet.get(i)).clone();
			cset.add(newCandidacy);
		}		
		
		// simplify candidacy set (remove candidacies which have equal identity component but are !=)

		ArrayList eliminatedCandidacies = new ArrayList();

		for (int i=0; i<cset.size(); i++)
		{
			for (int j=i+1; j<cset.size(); j++)
			{		
  			ArrayList candidacy1 = (ArrayList) cset.get(i);
	  		ArrayList candidacy2 = (ArrayList) cset.get(j);
	  		// if ids of candidacies match, and neither have already been eliminated, then remove 
	  		// the "lesser" candidacy
	  		if ( (!eliminatedCandidacies.contains(candidacy1)) &&
				     (!eliminatedCandidacies.contains(candidacy2)) &&
	 		   	   (((String)candidacy1.get(0)).equals((String) candidacy2.get(0))))
	  		{
	  			if (compare(candidacy1, candidacy2) >= 0)
	  				eliminatedCandidacies.add(candidacy2);
	  		}
			}
		}
		cset.removeAll(eliminatedCandidacies);
		
		return new ExtendedName(idset, Math.max(name1.splitNumber, name2.splitNumber), cset);		
	}
	
	/**
	 * Compare two candidacies (according to "Definition 2") c1 and c2, return an int less than,
	 * equal to, or greater than 0 if c1 is found, respectively, to be less than, equal to, or 
	 * greater than c2.
	 * (i.e., works like strcmp!).
	 * PRECONDITION: candidacies have same "i" component (i.e., same string in first element of candidacy)
	 * TBD: based on some assumptions.  Check with Dr. Boloni for rule accuracy.
	 * 
	 * @param candidacy1 first operand for comparison
	 * @param candidacy2 second operand for comparison
	 * @return -int (iff candidacy1<candidacy2); 0 (iff candidacy1==candidacy2); +int (iff candidacy1>candidacy2)
	 */
	private static int compare(ArrayList candidacy1, ArrayList candidacy2)
	{
		int split1 = ((Integer) candidacy1.get(1)).intValue();
		int split2 = ((Integer) candidacy2.get(1)).intValue();
		
		if (split1 < split2) 
			return -1;
		else if (split1 > split2) 
			return 1;
		else if (candidacy1.size() < candidacy2.size())
			return -1;
		else if (candidacy1.size() > candidacy2.size())
			return 1;
		// cannot recurse any longer... candidacies must be equal
		else if (candidacy1.size() == 2)
		  return 0;
		else
		{
			// clone, to make up for the fact that we're not passing arraylists by value when recursing
			ArrayList c1 = (ArrayList) candidacy1.clone();
			ArrayList c2 = (ArrayList) candidacy2.clone();
			c1.remove(1);
			c2.remove(1);
			return compare(c1, c2);
		}
	}
	
	/**
	 * Returns the next alpha in the sequence.  Will add a digit at the front of curAlpha if req'd.
	 * 
	 * @param curAlpha the current code
	 * @return the generated alpha code (i.e., curAlpha++)
	 */
	private static String getNextIdentity()
	{
		if ("".equals(nextIdentity))
		{ 
			nextIdentity = "A";
		  return "A";
		}
		
		boolean addDigit = false;
		char[] chs = nextIdentity.toCharArray();
    
		for (int i=chs.length - 1; i>=0; i--)
		{
			char ch = chs[i];
			// increment letter or number
			if (ch >= 65 && ch < 90)
			{
				ch++;
				chs[i] = ch;
				break;
			}
			// current position char resets to A, move to next position
			else if (i > 0)
			{
				chs[i]=65;
			}
			else
			{
				chs[i]=65;
				addDigit = true;
				break;
			}
		} 
		
		String next = new String(chs);
		if (addDigit)
		{
			next = "A" + next;
		}
		
		nextIdentity = next;
		return nextIdentity;
	}	
	
	// for testing.
	// TBD: almost there.  I get M1 (merge gc1+gc3): ({A,B}, 2, {(A,0)})
	//                               Dr Boloni gets: ({A,B}, 2, {})
	//       Is there a rule I'm not seeing which says empty cand > full cand ?
	public static void main(String args[])
	{
		ExtendedName en = new ExtendedName();
		System.out.println("parent: " + en);
		
		ArrayList list = en.split();
		ExtendedName en1 = (ExtendedName) list.get(0);
		ExtendedName en2 = (ExtendedName) list.get(1);
		System.out.println("child 1: " + en1);
		System.out.println("child 2: " + en2);

		System.out.println(" now splitting child 1...");
		list = en1.split();
		ExtendedName en1_1 = (ExtendedName) list.get(0);
		ExtendedName en1_2 = (ExtendedName) list.get(1);
		System.out.println("grandchild 1: " + en1_1);
		System.out.println("grandchild 2: " + en1_2);						
		
		System.out.println(" now splitting child 2...");
		list = en2.split();
		ExtendedName en2_1 = (ExtendedName) list.get(0);
		ExtendedName en2_2 = (ExtendedName) list.get(1);
		System.out.println("grandchild 3: " + en2_1);
		System.out.println("grandchild 4: " + en2_2);				
		
		ExtendedName m1 = merge(en1_1, en2_1);
		ExtendedName m2 = merge(en1_2, en2_2);
		System.out.println("merge gc1+gc3: " + m1);
		System.out.println("merge gc2+gc4: " + m2);
		
		// should be 0
		System.out.println(compare((ArrayList) en1_2.candidacySet.get(0), (ArrayList) en1_2.candidacySet.get(0)));
		
		// should be 1
		System.out.println(compare((ArrayList) en1_2.candidacySet.get(0), (ArrayList) en2_1.candidacySet.get(0)));
		
		// should be -1
		System.out.println(compare((ArrayList) en2_1.candidacySet.get(0), (ArrayList) en1_2.candidacySet.get(0)));
		
		// should be -1: (A,0) < (A,0,0)
		System.out.println(compare((ArrayList) en2_1.candidacySet.get(0), (ArrayList) en2_2.candidacySet.get(1)));
		
		// should be 1: (A,1) > (A,0,0)
		System.out.println(compare((ArrayList) en1_2.candidacySet.get(0), (ArrayList) en2_2.candidacySet.get(1)));		
		
	}

}
