package yaes.application.uavaltclient.nav;

import javax.vecmath.Vector2d;
import javax.vecmath.Vector3d;

import yaes.application.uavaltclient.util.Props;


/**
 * Performs social potential field calcualations for the purpose of keeping aircraft in 
 * "formation." 
 * TBD: attractive force tends to degrade too quickly.  Will probably need to tune.
 * 
 * @author dejongp
 */
public class SocPotField
{
	// Soc potential field constants.  These are undefined values which cause the point of
	// equilibrium (where attraction==repulsion) to change.
	// Rules: C1, C2 >= 0 ; SIGMA1 > SIGMA2 > 0

  // (these are currently the exact values used in [Reif 179])
	private static final double C1     = Props.getDouble("C1");
	private static final double C2     = Props.getDouble("C2");
	private static final double SIGMA1 = Props.getDouble("SIGMA1");
	private static final double SIGMA2 = Props.getDouble("SIGMA2");
	
	// easiest way to scale point of equilibrium.  r /= R_SCALE.
	private static final double R_SCALE  = Props.getDouble("R_SCALE"); //P.O.E. =~ 7 sim coordinate units
	
	// indicates force is significant enough to act on (more important than mission)
	public static final double SIGNFICANT_FORCE = Props.getDouble("SIGNFICANT_FORCE");
	
  // results after computations
	private double force; //force 2d force (no units)
	private double hdg; //hdg 2d heading (deg)
	private double yForce; //force in alt direction (no units)

		
	/**
	 * Perform the social potential field calculations for an aircraft, given their current position 
	 * (Pos) and the positions of other aircraft exerting attraction/repulsion on the aircraft.
	 * For 3d vectors I am using simulations coordinates (x=E-W, y=UP-DOWN, z=N-S).  For 2d vectors I 
	 * am using conventional coordinates: (x=E-W, y=N-S).
	 * 
	 * @param refAircraftPos point of reference (origin) for calculations. This is aircraft position of UAV making the calculations. 
	 * @param positions array of aircraft positions
	 */
	public void computeField(Pos refAircraftPos, Pos[] positions)
	{
		Vector3d refAircraft = new Vector3d(refAircraftPos.getX(), refAircraftPos.getY(), refAircraftPos.getZ());
		Vector3d resultant = new Vector3d();
		
		for(int i=0; i<positions.length; i++)
		{
			// "subtract off" refAircraft, since refAircraft is considered origin for vector calculations
			Vector3d v = new Vector3d(
				positions[i].getX() - refAircraft.x,
				positions[i].getY() - refAircraft.y,
				positions[i].getZ() - refAircraft.z);
				
		  // perform soc pot field calculation:
		  //     f(r) = - C1/r^SIG1 + C2/r^SIG2
		  double r = v.length();
		  
		  // scaling of r is easiest way to tune equilibrium
		  r /= R_SCALE;
		  
		  double localForce = - (C1 / Math.pow(r, SIGMA1)) + (C2 / Math.pow(r, SIGMA2));
		  				
		  // normalize, scale vector to the newly computed force, and add this computed vector to the 
		  // resultant vector		  
		  v.normalize();
			resultant.scaleAdd(localForce, v, resultant); 
		}
			
	  // resultant length is 3d force... we want 2d force.
	  // TBD: is this wack?  maybe we should be using 2d vectors all along.  But this way gives me 
	  // altitude difference.
	  Vector2d force2dVector = new Vector2d(resultant.x, resultant.z);
		//System.out.println("2d: " + force2dVector.x + " " + force2dVector.y);
		
		//adjust 180 deg. for sim coordinate system
		///NOT SURE WHY .angle() DIDN'T WORK 
		///double tmpHdg = Math.toDegrees(NavUtil.VEC_DUE_NORTH.angle(force2dVector));
		double tmpHdg = NavUtil.heading(0.0, 0.0, force2dVector.y, force2dVector.x);
		
		//System.out.println("tmpHdg:" + tmpHdg + " tmpHdg2:" + tmpHdg2);
		
		if (tmpHdg >= 180.0)
			tmpHdg -= 180.0;
	  else
			tmpHdg += 180.0;
			
		//System.out.println("tmpHdg flip:" + tmpHdg);
	
		
	  // populate return parameters
	  force = force2dVector.length();
	  hdg = tmpHdg;
	  yForce = resultant.y;
	}
	
	
  /**
   * @return magnatude of force
   */
  public double getForce()
  {
    return force;
  }

  /**
   * Obtain net direction the force is pushing the reference aircraft.
   * 
   * @return heading the aircraft is being forced (direction of force)
   */
  public double getHdg()
  {
    return hdg;
  }

  /**
   * TBD: currently unused.  Not sure if I will be using 3d soc pot fields.
   * Obtain 'Y' force component (up/down in sim coordinates)
   * 
   * @return Y force magnitude and direction (- indicates [TBD: up?down?])
   */
  public double getYForce()
  {
    return yForce;
  }

  /**
   * Return string form of soc pot field data.
   * 
   * @return string form of soc pot field data
   */
  public String getString()
  {
  	return "force:" + force + " hdg:" + hdg + " yforce:" + yForce;
  }
  
  
	/**
	 * For testing only.
	 * 
	 * @param args
	 */
	public static void main(String args[])
	{
		SocPotField spf = new SocPotField();
  	
		// from desktop\spfout.txt...
		// line 40 uav2
		Pos currentPos = new Pos(99.42, 157.0, 80.66); //uav2
		Pos[] positions = new Pos[1];
		positions[0] = new Pos(78.95, 151.0, 77.68); //uav1
		spf.computeField(currentPos, positions);
		System.out.println(spf.getString() );
		
		// line 44 uav1
		currentPos = new Pos(79.95, 151.0, 77.99); //uav1
		positions = new Pos[1];
		positions[0] = new Pos(97.32, 157.0, 80.0); //uav2
		spf.computeField(currentPos, positions);
		System.out.println(spf.getString() );		
		
		/* 
		Pos currentPos = new Pos(0.0, 0.0, 0.0);  
  	
		Pos[] positions = new Pos[1];
		int i=0;
		for (double d=0.1; d<50.0; d+=0.1)
		{
			positions[0] = new Pos(d, d, d); 	
			spf.computeField(currentPos, positions);
			System.out.println(++i + "\tFORCE:" + spf.getForce() + " HDG:" + spf.getHdg() + " dALT:" + spf.getYForce() + " (d=" + d + ")");
		}

		System.out.println("\n\nPOS=2\n\n");

		i=0;
		positions = new Pos[2];
		for (double d=0.1; d<50.0; d+=0.1)
		{
			positions[0] = new Pos(d, d, d);
			positions[1] = new Pos(d+d, d, d-(.5*d)); 	
			spf.computeField(currentPos, positions);
			System.out.println(++i + "\tFORCE:" + spf.getForce() + " HDG:" + spf.getHdg() + " dALT:" + spf.getYForce() + " (d=" + d + ")");
		}
		*/
	}
}
