package yaes.application.uavaltclient.simulation;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.net.Socket;

import yaes.application.uavaltclient.agents.UAVType;
import yaes.application.uavaltclient.missionmgr.UAVMap;
import yaes.application.uavaltclient.nav.Pos;


/**
 * A proxy to the simulation server.  Exposes required sim functionality.  Communicates with
 * Simulation Server.  Each agent is expected to maintain its own SimProxy.
 * 
 * @author dejongp
 */
public class SimProxy
{
	// connection constants
	private static final String SIM_SERVER_IP = "localhost";
	private static final int SIM_SERVER_PORT = 20100;
	
	// message id constants
	private static final int INIT_MSG = 0;
	private static final int MOVE_MSG = 1;
	private static final int POSITION_MSG = 3;
	private static final int HEIGHT_MSG = 4;
	private static final int ENEMY_MSG = 5;
	private static final int DAMAGE_MSG = 7;
	private static final int ANGLE_MSG = 9;
	private static final int VELOCITY_MSG = 10;
	private static final int CLOSE_MSG = 11;
	private static final int FUEL_MSG = 12;
	private static final int DISTANCE_MSG = 13;
	private static final int DROP_BOMB_MSG = 14;
	private static final int TEAM_COMMUNICATE_MSG = 15;
	private static final int UNIT_COMMUNICATE_MSG = 16;
	private static final int TEAM_MSG = 17;
	private static final int LINE_OF_SIGHT_MSG = 18;
	private static final int TEAM_MEMBERS_MSG = 19;
	private static final int RADAR_RADIUS_MSG = 20;
	private static final int HEIGHT_MAP_MSG = 21;
	private static final int RANGE_FINDER_MSG = 22;
	
	// uav types
	private static final int UAV_SIMTYPE_NORMAL = 0;
	private static final int UAV_SIMTYPE_BOMBER = 1;
	private static final int UAV_SIMTYPE_SCOUT = 2;
	private static final int UAV_SIMTYPE_FIGHTER = 3;
	
	
	private BufferedReader connectionIn;
	private PrintWriter connectionOut;
	
	/**
	 * 
	 * @throws SimException
	 */
	public SimProxy() throws SimException
	{
		try
		{
			Socket server = new Socket(SIM_SERVER_IP, SIM_SERVER_PORT);
      
			InputStreamReader isr = new InputStreamReader(server.getInputStream() );
			connectionIn = new BufferedReader(isr);
			connectionOut = new PrintWriter(server.getOutputStream(), true);
		}
		catch(IOException e)
		{
			throw new SimException(e);
		}
	}
	
	/**
	 * Initializes the agent with a name and a type (Normal, Scout, Bomber).
	 * 
	 * @param name Identifier for this UAV
	 * @param type UAV type (from UAVType)
	 * @throws SimException
	 */
	public void init(String name, String type) throws SimException
	{
		int uavType;
		
		if (UAVType.UAV_TYPE_BOMBER.equals(type))
			uavType = UAV_SIMTYPE_BOMBER;
		else if (UAVType.UAV_TYPE_NORMAL.equals(type))
			uavType = UAV_SIMTYPE_NORMAL;
		else if (UAVType.UAV_TYPE_SCOUT.equals(type))
			uavType = UAV_SIMTYPE_SCOUT;		
		else
			throw new SimException("invalid type:" + type);
		
		send(INIT_MSG + " " + name + " " + uavType);
	}

  /**
   * Instruct a UAV to be moved within the simulation.
   * 
   * @param velocity
   * @param height lowering height means gaining altitude (note: height at start is approx 129.0)
   * @param orientation positive orientation means clockwise rotation
   */
  public void move(double velocity, double height, double orientation) throws SimException
  {
  	send(MOVE_MSG + " " + velocity + " " + height + " " + orientation);
  }
  
  /**
   * Obtain the current x,y,z (z=MSL) position from the simulation.
   * 
   * @return
   * @throws SimException
   */
  public Pos position() throws SimException
  {
  	String posString = send(Integer.toString(POSITION_MSG));
  	String sarray[] = posString.split(" ");
  	Pos pos = new Pos(
  	  Double.parseDouble(sarray[1]),
		  Double.parseDouble(sarray[2]),
		  Double.parseDouble(sarray[3]));
		
  	return pos;
  }
  
  /**
   * Obtain the current AGL height above ground of the uav from the simulation.
   * 
   * @return
   * @throws SimException
   */
  public double height() throws SimException
  {
		String heightString = send(Integer.toString(HEIGHT_MSG));
		String sarray[] = heightString.split(" ");
		return Double.parseDouble(sarray[1]);
  }

  /**
   * Obtain angle from absolute north of the uav from the simulation.
   * UNITS: DEGREES
   * 
   * @return angle in degrees
   * @throws SimException
   */
  public double angle() throws SimException
  {
		String angleString = send(Integer.toString(ANGLE_MSG));
		String sarray[] = angleString.split(" ");
		return Double.parseDouble(sarray[1]);  	
  }
  
  /**
   * Obtain velocity of the uav from the simulation.
   * 
   * @return
   * @throws SimException
   */
  public double velocity() throws SimException
  {
    String velString = send(Integer.toString(VELOCITY_MSG));
		String sarray[] = velString.split(" ");
		return Double.parseDouble(sarray[1]); 	
  }

  /**
   * Assign this uav to a team.
   * 
   * @param teamID
   * @throws SimException
   */
	public void team(int teamID) throws SimException
	{
		send(TEAM_MSG + " " + teamID);
	}
	
	/**
	 * Scout enemies which have been sighted.
	 * 
	 * @return Pos if enemy is spotted, the position of the enemy is returned.  Otherwise, null.
	 */
	public Pos enemy() throws SimException
	{
		String str = send(Integer.toString(ENEMY_MSG));
		String sarray[] = str.split(" ");
		Pos pos = null;
		
		if (sarray.length > 2)
		  pos = new Pos(Double.parseDouble(sarray[2]), Double.parseDouble(sarray[3]), Double.parseDouble(sarray[4]));
		
		return pos;
	}
	
	/**
	 * Drop a bomb at the present location.
	 * 
	 * @throws SimException
	 */
	public void dropBomb() throws SimException
	{
		send(Integer.toString(DROP_BOMB_MSG));
	}
	
	/**
	 * Updates UAV map with a window of height values surrounding the uav (for efficiency).
	 * 
	 * @param map the UAVMap containing the height map to update
	 * @param stepSize for stepSize X, return every Xth coordinate's height in the window
	 * @param aircraftHeight, used to determine absolute position of obstacles rather than relative to aircraft
	 * @throws SimException
	 */
	public void rangeFinder(UAVMap map, int stepSize, int aircraftHeight) throws SimException
	{
		String gridstr = send(RANGE_FINDER_MSG + " " + stepSize);
		
		///TODO: tmp
		///System.out.println("SENT: " + RANGE_FINDER_MSG + " " + stepSize);
		
		//System.out.println("GOT: " + gridstr);
		
		String sarray[] = gridstr.split(" ");
		
		int step = Integer.parseInt(sarray[1]);   // should be same as stepSize
		int windowLength = Integer.parseInt(sarray[2]); // (window/step) = each dimension of grid
		//int y = Integer.parseInt(sarray[3]);      // (not sure how y is useful)
		int startX = Integer.parseInt(sarray[4]); // top left X coordinate of the window
		int startZ = Integer.parseInt(sarray[5]); // top left Z coordinate of the window

		///System.out.println("startX,startZ : " + startX + "," + startZ);

		// computations are only accurate for positive X and Z
		if ((startX > 0) && (startZ > 0))
		{
			int rowLength = windowLength / step;
			
			int hmap[][] = map.getHeightMap();
			int currentXArray = 0;
      
			// ignore first 6 elements containing metadata
			for (int i = 6; i < sarray.length; i++)
			{
				int i_adj = i - 6;
				
				if (i_adj % rowLength == 0)
				{
					currentXArray += step;
					///System.out.println();
				}

				int x = currentXArray + startX;
				int z = ((i_adj % rowLength) * step) + startZ;

				///System.out.print(x + "," + z + "-" + (aircraftHeight - Integer.parseInt(sarray[i])) + " ");
				try
			  {
			  	hmap[x][z] = aircraftHeight - Integer.parseInt(sarray[i]);
			  }
			  catch(ArrayIndexOutOfBoundsException e)
			  {
			    System.out.println(e + " at x,z " + x + "," + z);
			  }
			}
		}
	}
	
	/**
	 * Sends a synchronous request to simulation server, returns response received to caller.
	 * 
	 * @param msg Message to send to simulation server
	 * @return response received from simulation server
	 * @throws SimException when no response received from sim server
	 */
	private String send(String msg) throws SimException
	{
		try
		{
			connectionOut.println(msg);
			String response = connectionIn.readLine();
			if (response == null)
	  	{
	  		throw new SimException("No response from Simulation Server");
	  	}
	  	return response;
		}
		catch(IOException e)
		{
			throw new SimException(e);
		}
	}
	
	/**
	 * test driver
	 */
	public static void main(String args[]) throws Exception 
	{
		
		int RESET = 50; //iterations
		
		UAVMap map = new UAVMap();
		SimProxy sim = new SimProxy();
		sim.init("bif", UAVType.UAV_TYPE_BOMBER);
		sim.team(0);
		sim.move(1.2, 205, 12);
		int cnt = 0;
		for (;;)
		{
			cnt++;
			Thread.sleep(100);
			Pos pos = sim.position();
			double height = sim.height();
			double angle = sim.angle();
			double vel = sim.velocity();
			System.out.println("#:" + cnt + " x:" + pos.getX() + " y:" + pos.getY() + " z:" + pos.getZ() + 
				" hgt:" + height + " ang:" + angle + " v:" + vel);
									
			if ((cnt < (RESET / 2) || height < 20.0))
			  sim.move(1.2, 10, 0.2);
			else
			  sim.move(1.2, -10, -0.2);
			
			if (cnt % 25 == 0)
			{
				sim.rangeFinder(map, 4, 100);
				//map.printMap();
			}
			
			if (cnt == RESET) cnt = 0;
		}
		

	}
}
